# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module contains the abstract base classes that define Strawberry Fields
compatible statevector simulator backends."""

# pylint: disable=no-self-use,missing-docstring


class NotApplicableError(TypeError):
    """Exception raised by the backend when the user attempts an unsupported operation.
    E.g. :meth:`measure_fock` on a Gaussian backend.
    Conceptually different from NotImplementedError (which means "not implemented, but at some point may be").
    """


class ModeMap:
    """
    Simple internal class for maintaining a map of existing modes.
    """

    def __init__(self, num_subsystems):
        self._init = num_subsystems
        #: list[int]: _map[k] is the internal index used by the backend for
        # computational mode k, or None if the mode has been deleted
        self._map = [k for k in range(num_subsystems)]

    def reset(self):
        """reset the modemap to the initial state"""
        self._map = [k for k in range(self._init)]

    def _single_mode_valid(self, mode):
        if mode is None:
            return False

        if 0 <= mode < len(self._map):
            return True

        return False

    def _reduce_to_existing_modes(self, modes):
        # Reduces modes to only those which are not None in the map
        if isinstance(modes, int):
            modes = [modes]
        return [m for m in modes if m in self._map]

    def remap(self, modes):
        """Remaps the mode list"""
        if isinstance(modes, int):
            modes = [modes]
            was_int = True
        else:
            was_int = False

        modes_list = [self._map[m] for m in modes]

        if was_int:
            return modes_list[0]

        return modes_list

    def valid(self, modes):
        """checks if the mode list is valid"""
        if modes is None:
            return False

        if isinstance(modes, int):
            modes = [modes]

        # pylint: disable=len-as-condition
        if len(modes) == 0 or len(modes) > len(self._map):
            return False

        for m in modes:
            if not self._single_mode_valid(m):
                return False

        return True

    def show(self):
        """Returns the mapping"""
        return self._map

    def delete(self, modes):
        """Deletes a mode"""
        if isinstance(modes, int):
            modes = [modes]
        if self.valid(modes):
            new_map = []
            ctr = 0
            for m in range(len(self._map)):
                if m in modes or self._map[m] is None:
                    new_map.append(None)
                else:
                    new_map.append(ctr)
                    ctr += 1
            self._map = new_map
        else:
            raise ValueError("Specified modes for deleting are invalid.")

    def add(self, num_modes):
        """Adds a mode"""
        num_active_modes = len([m for m in self._map if m is not None])
        self._map += [k for k in range(num_active_modes, num_active_modes + num_modes)]


class BaseBackend:
    """Abstract base class for backends."""

    #: str: short name of the backend
    short_name = "base"
    #: str, None: Short name of the CircuitSpecs class used to validate Programs for this backend. None if no validation is required.
    circuit_spec = None

    def __init__(self):
        self._supported = {}

    def __str__(self):
        """String representation."""
        return self.__class__.__name__

    def supports(self, name):
        """Check whether the backend supports the given operating mode.

        Currently supported operating modes are:

        * "gaussian": for manipulations in the Gaussian representation using the
          displacements and covariance matrices
        * "fock_basis": for manipulations in the Fock representation
        * "mixed_states": for representations where the quantum state is mixed
        * "batched": allows for a multiple circuits to be simulated in parallel

        Args:
            name (str): name of the operating mode which we are checking support for

        Returns:
            bool: True if this backend supports that operating mode.
        """
        return self._supported.get(name, False)

    def begin_circuit(self, num_subsystems, **kwargs):
        r"""Instantiate a quantum circuit.

        Instantiates a representation of a quantum optical state with ``num_subsystems`` modes.
        The state is initialized to vacuum.

        The modes in the circuit are indexed sequentially using integers, starting from zero.
        Once an index is assigned to a mode, it can never be re-assigned to another mode.
        If the mode is deleted its index becomes invalid.
        An operation acting on an invalid or unassigned mode index raises an ``IndexError`` exception.

        Args:
            num_subsystems (int): number of modes in the circuit

        Keyword Args:
            cutoff_dim (int): Hilbert space truncation dimension (for Fock basis backends only)
            batch_size (int): (optional) batch-axis dimension, enables batched operation if > 1 (for the TF backend only)
        """
        # BaseBackend can be instantiated for testing purposes, even though it does not do anything.

    def add_mode(self, n=1):
        """Add modes to the circuit.

        The new modes are initialized to the vacuum state.
        They are assigned mode indices sequentially, starting from the first unassigned index.

        Args:
            n (int): number of modes to add

        Returns:
            list[int]: indices of the newly added modes
        """
        raise NotImplementedError

    def del_mode(self, modes):
        """Delete modes from the circuit.

        The deleted modes are traced out.
        As a result the state may have to be described using a density matrix.

        The indices of the deleted modes become invalid for the lifetime of the circuit object.
        They will never be reassigned to other modes.
        Deleting a mode that has already been deleted raises an ``IndexError`` exception.

        Args:
            modes (Sequence[int]): mode numbers to delete
        """
        raise NotImplementedError

    def get_modes(self):
        """Return a list of the active modes for the circuit.

        A mode is active if it has been created and has not been deleted.

        Returns:
            list[int]: sorted list of active (assigned, not invalid) mode indices
        """
        raise NotImplementedError

    def reset(self, pure=True, **kwargs):
        """Reset the circuit so that all the modes are in the vacuum state.

        After the reset the circuit is in the same state as it was after
        the last :meth:`begin_circuit` call. It will have the original number
        of modes, all initialized in the vacuum state. Some circuit parameters
        may be changed during the reset, see the keyword args below.

        Args:
            pure (bool): if True, initialize the circuit in a pure state representation
                (will use a mixed state representation if pure is False)

        Keyword Args:
            cutoff_dim (int): new Hilbert space truncation dimension (for Fock basis backends only)
        """
        raise NotImplementedError

    def prepare_vacuum_state(self, mode):
        """Prepare the vacuum state in the specified mode.

        The requested mode is traced out and replaced with the vacuum state.
        As a result the state may have to be described using a density matrix.

        Args:
            mode (int): which mode to prepare the vacuum state in
        """
        raise NotImplementedError

    def prepare_coherent_state(self, alpha, mode):
        r"""Prepare a coherent state in the specified mode.

        The requested mode is traced out and replaced with the coherent state :math:`\ket{\alpha}`.
        As a result the state may have to be described using a density matrix.

        Args:
            alpha (complex): coherent state displacement parameter
            mode (int): which mode to prepare the coherent state in
        """
        raise NotImplementedError

    def prepare_squeezed_state(self, r, phi, mode):
        r"""Prepare a squeezed vacuum state in the specified mode.

        The requested mode is traced out and replaced with the squeezed state :math:`\ket{z}`,
        where :math:`z=re^{i\phi}`.
        As a result the state may have to be described using a density matrix.

        Args:
            r (float): squeezing amplitude
            phi (float): squeezing angle
            mode (int): which mode to prepare the squeezed state in
        """
        raise NotImplementedError

    def prepare_displaced_squeezed_state(self, alpha, r, phi, mode):
        r"""Prepare a displaced squeezed state in the specified mode.

        The requested mode is traced out and replaced with the displaced squeezed
        state state :math:`\ket{\alpha, z}`, where :math:`z=re^{i\phi}`.
        As a result the state may have to be described using a density matrix.

        Args:
            alpha (complex): displacement parameter
            r (float): squeezing amplitude
            phi (float): squeezing angle
            mode (int): which mode to prepare the squeezed state in
        """
        raise NotImplementedError

    def prepare_thermal_state(self, nbar, mode):
        r"""Prepare a thermal state in the specified mode.

        The requested mode is traced out and replaced with the thermal state :math:`\rho(nbar)`.
        As a result the state may have to be described using a density matrix.

        Args:
            nbar (float): thermal population (mean photon number) of the mode
            mode (int): which mode to prepare the thermal state in
        """
        raise NotImplementedError

    def rotation(self, phi, mode):
        """Apply the phase-space rotation operation to the specified mode.

        Args:
            phi (float): rotation angle
            mode (int): which mode to apply the rotation to
        """
        raise NotImplementedError

    def displacement(self, alpha, mode):
        """Apply the displacement operation to the specified mode.

        Args:
            alpha (complex): displacement parameter
            mode (int): which mode to apply the displacement to
        """
        raise NotImplementedError

    def squeeze(self, z, mode):
        """Apply the squeezing operation to the specified mode.

        Args:
            z (complex): squeezing parameter
            mode (int): which mode to apply the squeeze to
        """
        raise NotImplementedError

    def beamsplitter(self, t, r, mode1, mode2):
        """Apply the beamsplitter operation to the specified modes.

        It is assumed that :math:`|r|^2+|t|^2 = t^2+|r|^2=1`, i.e that t is real.

        Args:
            t (float): transmitted amplitude
            r (complex): reflected amplitude (with phase)
            mode1 (int): first mode that beamsplitter acts on
            mode2 (int): second mode that beamsplitter acts on
        """
        raise NotImplementedError

    def loss(self, T, mode):
        r"""Perform a loss channel operation on the specified mode.

        Args:
            T (float): loss parameter, :math:`0\leq T\leq 1`.
            mode (int): index of mode where operation is carried out
        """
        raise NotImplementedError

    def thermal_loss(self, T, nbar, mode):
        r"""Perform a thermal loss channel operation on the specified mode.

        Args:
            T (float): loss parameter, :math:`0\leq T\leq 1`.
            nbar (float): mean photon number of the environment thermal state
            mode (int): index of mode where operation is carried out
        """
        raise NotImplementedError

    def measure_homodyne(self, phi, mode, shots=1, select=None, **kwargs):
        r"""Measure a :ref:`phase space quadrature <homodyne>` of the given mode.

        For the measured mode, samples the probability distribution
        :math:`f(q) = \bra{q_\phi} \rho \ket{q_\phi}`
        and returns the sampled value.
        Here :math:`\ket{q_\phi}` is the eigenstate of the operator

        .. math::
           \hat{q}_\phi = \sqrt{2/\hbar}(\cos(\phi)\hat{x} +\sin(\phi)\hat{p}) = e^{-i\phi} \hat{a} +e^{i\phi} \hat{a}^\dagger.

        .. note::
           This method is :math:`\hbar` independent.
           The returned values can be converted to conventional position/momentum
           eigenvalues by multiplying them with :math:`\sqrt{\hbar/2}`.

        Updates the current state such that the measured mode is reset
        to the vacuum state. This is because we cannot represent exact position or
        momentum eigenstates in any of the backends, and experimentally the photons
        are destroyed in a homodyne measurement.

        Args:
            phi (float): phase angle of the quadrature to measure (x: :math:`\phi=0`, p: :math:`\phi=\pi/2`)
            mode (int): which mode to measure
            shots (int): number of measurement samples to obtain
            select (None or float): If not None: desired value of the measurement result.
                Enables post-selection on specific measurement results instead of random sampling.

        Keyword arguments can be used to pass additional parameters to the backend.
        Options for such arguments will be documented in the respective subclasses.

        Returns:
            float: measured value
        """
        raise NotImplementedError

    def measure_fock(self, modes, shots=1, select=None, **kwargs):
        """Measure the given modes in the Fock basis.

        ..note::
          When :code:``shots == 1``, updates the current system state to the conditional state of that
          measurement result. When :code:``shots > 1``, the system state is not updated.

        Args:
            modes (Sequence[int]): which modes to measure
            shots (int): number of measurement samples to obtain
            select (None or Sequence[int]): If not None: desired values of the measurement results.
                Enables post-selection on specific measurement results instead of random sampling.
                ``len(select) == len(modes)`` is required.
        Returns:
            tuple[int]: measurement results
        """
        raise NotImplementedError

    def measure_threshold(self, modes, shots=1, select=None, **kwargs):
        """Measure the given modes in the thresholded Fock basis, i.e., zero or nonzero photons).

        .. note::

            When :code:``shots == 1``, updates the current system state to the conditional state of that
            measurement result. When :code:``shots > 1``, the system state is not updated.

        Args:
            modes (Sequence[int]): which modes to measure
            shots (int): number of measurement samples to obtain
            select (None or Sequence[int]): If not None: desired values of the measurement results.
                Enables post-selection on specific measurement results instead of random sampling.
                ``len(select) == len(modes)`` is required.
        Returns:
            tuple[int]: measurement results
        """
        raise NotImplementedError

    def is_vacuum(self, tol=0.0, **kwargs):
        r"""Test whether the current circuit state is vacuum (up to given tolerance).

        Returns True iff :math:`|\bra{0} \rho \ket{0} -1| \le` ``tol``, i.e.,
        the fidelity of the current circuit state with the vacuum state is within
        the given tolerance from 1.

        Args:
            tol (float): numerical tolerance

        Returns:
            bool: True iff current state is vacuum up to tolerance tol
        """
        raise NotImplementedError

    def state(self, modes=None, **kwargs):
        r"""Returns the state of the quantum simulation.

        Args:
            modes (int or Sequence[int] or None): Specifies the modes to restrict the return state to.
                None returns the state containing all the modes.
                The returned state contains the requested modes in the given order, i.e.,
                ``modes=[3,0]`` results in a two mode state being returned with the first mode being
                subsystem 3 and the second mode being subsystem 0.
        Returns:
            BaseState: state description, specific child class depends on the backend
        """
        raise NotImplementedError


# =============================
# Fock-basis backends
# =============================


class BaseFock(BaseBackend):
    """Abstract base class for backends capable of Fock state manipulation."""

    def __init__(self):
        super().__init__()
        self._supported["fock_basis"] = True

    def get_cutoff_dim(self):
        """Returns the Hilbert space cutoff dimension used.

        Returns:
            int: cutoff dimension
        """
        raise NotImplementedError

    def prepare_fock_state(self, n, mode):
        r"""Prepare a Fock state in the specified mode.

        The requested mode is traced out and replaced with the Fock state :math:`\ket{n}`.
        As a result the state may have to be described using a density matrix.

        Args:
            n (int): Fock state to prepare
            mode (int): which mode to prepare the Fock state in
        """
        raise NotImplementedError

    def prepare_ket_state(self, state, modes):
        r"""Prepare the given ket state in the specified modes.

        The requested modes are traced out and replaced with the given ket state
        (in the Fock basis). As a result the state may have to be described using a
        density matrix.

        Args:
            state (array): Ket state in the Fock basis.
                The state can be given in either vector form, with one index,
                or tensor form, with one index per mode. For backends supporting batched
                mode, state can be a batch of such vectors or tensors.
            modes (int or Sequence[int]): Modes to prepare the state in.
                If modes is not ordered this is taken into account when preparing the state,
                i.e., when a two mode state is prepared in modes=[3,1], then the first
                mode of state goes into mode 3 and the second mode goes into mode 1 of the simulator.
        """
        raise NotImplementedError

    def prepare_dm_state(self, state, modes):
        r"""Prepare the given mixed state in the specified modes.

        The requested modes are traced out and replaced with the given density matrix
        state (in the Fock basis).
        As a result the state will be described using a density matrix.

        Args:
            state (array): Density matrix in the Fock basis.
                The state can be given in either matrix form, with two indices, or tensor
                form, with two indices per mode. For backends supporting batched mode,
                state can be a batch of such matrices or tensors.
            modes (int or Sequence[int]): which mode to prepare the state in
                If modes is not ordered this is take into account when preparing the
                state, i.e., when a two mode state is prepared in modes=[3,1], then
                the first mode of state goes into mode 3 and the second mode goes
                into mode 1 of the simulator.
        """
        raise NotImplementedError

    def cubic_phase(self, gamma, mode):
        r"""Apply the cubic phase operation to the specified mode.

        Applies the operation

        .. math::
           \exp\left(i \frac{\gamma}{6} (\hat{a} +\hat{a}^\dagger)^3\right)

        to the specified mode.

        .. note::
           This method is :math:`\hbar` independent.
           The usual definition of the cubic phase gate is :math:`\hbar` dependent:

           .. math::
              V(\gamma') = \exp\left(i \frac{\gamma'}{3\hbar} \hat{x}^3\right) = \exp\left(i \frac{\gamma' \sqrt{\hbar/2}}{6} (\hat{a} +\hat{a}^\dagger)^3\right).

           Hence the cubic phase gate :math:`V(\gamma')` is executed on a backend by scaling the
           :math:`\gamma'` parameter by :math:`\sqrt{\hbar/2}` and then passing it to this method,
           much in the way the :math:`\hbar` dependent `X` and `Z` gates are implemented through the
           :math:`\hbar` independent :meth:`~BaseBackend.displacement` method.

        .. warning::
            The cubic phase gate can suffer heavily from numerical inaccuracies
            due to finite-dimensional cutoffs in the Fock basis. The gate
            implementation in Strawberry Fields is unitary, but it
            does not implement an exact cubic phase gate. The Kerr gate
            provides an alternative non-Gaussian gate.

        Args:
            gamma (float): scaled cubic phase shift, :math:`\gamma = \gamma' \sqrt{\hbar/2}`
            mode (int): which mode to apply it to
        """
        raise NotImplementedError

    def kerr_interaction(self, kappa, mode):
        r"""Apply the Kerr interaction :math:`\exp{(i\kappa \hat{n}^2)}` to the specified mode.

        Args:
            kappa (float): strength of the interaction
            mode (int): which mode to apply it to
        """
        raise NotImplementedError

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        r"""Apply the two mode cross-Kerr interaction :math:`\exp{(i\kappa \hat{n}_1\hat{n}_2)}` to the specified modes.

        Args:
            kappa (float): strength of the interaction
            mode1 (int): first mode that cross-Kerr interaction acts on
            mode2 (int): second mode that cross-Kerr interaction acts on
        """
        raise NotImplementedError

    def state(self, modes=None, **kwargs):
        r"""Returns the state of the quantum simulation.

        See :meth:`.BaseBackend.state`.

        Returns:
            BaseFockState: state description
        """
        raise NotImplementedError


# ==============================
# Gaussian-formulation backends
# ==============================


class BaseGaussian(BaseBackend):
    """Abstract base class for backends that are only capable of Gaussian state manipulation."""

    def __init__(self):
        super().__init__()
        self._supported["gaussian"] = True

    def measure_heterodyne(self, mode, shots=1, select=None):
        r"""Perform a heterodyne measurement on the given mode.

        Updates the current state of the circuit such that the measured mode is reset to the vacuum state.

        Args:
            mode (int): which mode to measure
            shots (int): number of measurement samples to obtain
            select (None or complex): If not None: desired value of the measurement result.
                Enables post-selection on specific measurement results instead of random sampling.

        Returns:
            complex: measured value
        """
        raise NotImplementedError

    def prepare_gaussian_state(self, r, V, modes):
        r"""Prepare a Gaussian state.

        The specified modes are traced out and replaced with a Gaussian state
        provided via a vector of means and a covariance matrix.

        .. note::
           This method is :math:`\hbar` independent.
           The input arrays are the means and covariance of the
           :math:`a+a^\dagger` and :math:`-i(a-a^\dagger)` operators.
           They are obtained by dividing the xp means by :math:`\sqrt{\hbar/2}`
           and the xp covariance by :math:`\hbar/2`.

        Args:
            r (array): vector of means in xp ordering
            V (array): covariance matrix in xp ordering
            modes (int or Sequence[int]): Which modes to prepare the state in.
                If the modes are not sorted, this is taken into account when preparing the state.
                I.e., when a two mode state is prepared with ``modes=[3,1]``, the first
                mode of the given state goes into mode 3 and the second mode goes into mode 1.
        """
        raise NotImplementedError

    def get_cutoff_dim(self):
        raise NotApplicableError

    def prepare_fock_state(self, n, mode):
        raise NotApplicableError

    def prepare_ket_state(self, state, mode):
        raise NotApplicableError

    def prepare_dm_state(self, state, mode):
        raise NotApplicableError

    def cubic_phase(self, gamma, mode):
        raise NotApplicableError

    def kerr_interaction(self, kappa, mode):
        raise NotApplicableError

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        raise NotApplicableError

    def state(self, modes=None, **kwargs):
        """Returns the state of the quantum simulation.

        See :meth:`.BaseBackend.state`.

        Returns:
            BaseGaussianState: state description
        """
        raise NotImplementedError



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The module docstring is in strawberryfields/circuitspecs/__init__.py
"""
**Module name:** :mod:`strawberryfields.circuitspecs.circuit_specs`
"""

from typing import List, Set, Dict, Union
import abc

import networkx as nx
import blackbird
from blackbird.utils import to_DiGraph

import strawberryfields.program_utils as pu


class CircuitSpecs(abc.ABC):
    """Abstract base class for describing circuit classes.

    This class stores information about :term:`classes of quantum circuits <circuit class>`.
    For some circuit classes (e.g, ones corresponding to physical hardware chips), the
    specifications can be quite rigid. For other classes, e.g., circuits supported by a particular
    simulator backend, the specifications can be more flexible and general.

    Key ingredients in a specification include: the primitive gates supported by the circuit class,
    the gates that can be decomposed to sequences of primitive gates, and the possible
    topology/connectivity restrictions.

    This information is used e.g., in :meth:`.Program.compile` for validation and compilation.
    """

    short_name = ""
    """str: short name of the circuit class"""

    @property
    @abc.abstractmethod
    def modes(self) -> Union[int, None]:
        """The number of modes supported by the circuit class.

        If the circuit class supports arbitrary number of modes, set this to 0.

        Returns:
            int: number of supported modes
        """

    @property
    @abc.abstractmethod
    def local(self) -> bool:
        """Whether the circuit class can be executed locally (i.e., within a simulator).

        Returns:
            bool: ``True`` if the circuit class supports local execution
        """

    @property
    @abc.abstractmethod
    def remote(self) -> bool:
        """Whether the circuit class supports remote execution.

        Returns:
            bool: ``True`` if the circuit class supports remote execution
        """

    @property
    @abc.abstractmethod
    def interactive(self) -> bool:
        """Whether the circuits in the class can be executed interactively, that is,
        the registers in the circuit are not reset between engine executions.

        Returns:
            bool: ``True`` if the circuit supports interactive use
        """

    @property
    @abc.abstractmethod
    def primitives(self) -> Set[str]:
        """The primitive set of quantum operations directly supported
        by the circuit class.

        Returns:
            set[str]: the names of the quantum primitives the circuit class supports
        """

    @property
    @abc.abstractmethod
    def decompositions(self) -> Dict[str, Dict]:
        """Quantum operations that are not quantum primitives for the
        circuit class, but are supported via specified decompositions.

        This should be of the form

        .. code-block:: python

            {'operation_name': {'option1': val, 'option2': val,...}}

        For each operation specified in the dictionary, the
        :meth:`.Operation.decompose` method will be called during
        :class:`.Program` compilation, with keyword arguments
        given by the dictionary value.

        Returns:
            dict[str, dict]: the quantum operations that are supported
            by the circuit class via decomposition
        """

    @property
    def parameter_ranges(self) -> Dict[str, List[List[float]]]:
        """Allowed parameter ranges for supported quantum operations.

        This property is optional.

        Returns:
            dict[str, list]: a dictionary mapping an allowed quantum operation
            to a nested list of the form ``[[p0_min, p0_max], [p1_min, p0_max], ...]``.
            where ``pi`` corresponds to the ``i`` th gate parameter
        """
        return dict()

    @property
    def graph(self):
        """The allowed circuit topologies or connectivity of the class, modelled as a directed
        acyclic graph.

        This property is optional; if arbitrary topologies are allowed in the circuit class,
        this will simply return ``None``.

        Returns:
            networkx.DiGraph: a directed acyclic graph
        """
        if self.circuit is None:
            return None

        # returned DAG has all parameters set to 0
        bb = blackbird.loads(self.circuit)

        if bb.is_template():
            params = bb.parameters
            kwargs = {p: 0 for p in params}

            # initialize the topology with all template
            # parameters set to zero
            topology = to_DiGraph(bb(**kwargs))

        else:
            topology = to_DiGraph(bb)

        return topology

    @property
    def circuit(self):
        """A rigid circuit template that defines this circuit specification.

        This property is optional. If arbitrary topologies are allowed in the circuit class,
        **do not define this property**. In such a case, it will simply return ``None``.

        If a backend device expects a specific template for the recieved Blackbird
        script, this method will return the serialized Blackbird circuit in string
        form.

        Returns:
            Union[str, None]: Blackbird program or template representing the circuit
        """
        return None

    def compile(self, seq, registers):
        """Class-specific circuit compilation method.

        If additional compilation logic is required, child classes can redefine this method.

        Args:
            seq (Sequence[Command]): quantum circuit to modify
            registers (Sequence[RegRefs]): quantum registers
        Returns:
            List[Command]: modified circuit
        Raises:
            CircuitError: the given circuit cannot be validated to belong to this circuit class
        """
        # registers is not used here, but may be used if the method is overwritten pylint: disable=unused-argument
        if self.graph is not None:
            # check topology
            DAG = pu.list_to_DAG(seq)

            # relabel the DAG nodes to integers, with attributes
            # specifying the operation name. This allows them to be
            # compared, rather than using Command objects.
            mapping = {i: n.op.__class__.__name__ for i, n in enumerate(DAG.nodes())}
            circuit = nx.convert_node_labels_to_integers(DAG)
            nx.set_node_attributes(circuit, mapping, name="name")

            def node_match(n1, n2):
                """Returns True if both nodes have the same name"""
                return n1["name"] == n2["name"]

            # check if topology matches
            if not nx.is_isomorphic(circuit, self.graph, node_match):
                # TODO: try and compile the program to match the topology
                # TODO: add support for parameter range matching/compilation
                raise pu.CircuitError(
                    "Program cannot be used with the CircuitSpec '{}' "
                    "due to incompatible topology.".format(self.short_name)
                )

        return seq

    def decompose(self, seq):
        """Recursively decompose all gates in a given sequence, as allowed
        by the circuit specification.

        This method follows the directives defined in the
        :attr:`~.CircuitSpecs.primitives` and :attr:`~.CircuitSpecs.decompositions`
        class attributes to determine whether a command should be decomposed.

        The order of precedence to determine whether decomposition
        should be applied is as follows.

        1. First, we check if the operation is in :attr:`~.CircuitSpecs.decompositions`.
           If not, decomposition is skipped, and the operation is applied
           as a primitive (if supported by the ``CircuitSpecs``).

        2. Next, we check if (a) the operation supports decomposition, and (b) if the user
           has explicitly requested no decomposition.

           - If both (a) and (b) are true, the operation is applied
             as a primitive (if supported by the ``CircuitSpecs``).

           - Otherwise, we attempt to decompose the operation by calling
             :meth:`~.Operation.decompose` recursively.

        Args:
            list[strawberryfields.program_utils.Command]: list of commands to
                be decomposed

        Returns:
            list[strawberryfields.program_utils.Command]: list of compiled commands
            for the circuit specification
        """
        compiled = []
        for cmd in seq:
            op_name = cmd.op.__class__.__name__
            if op_name in self.decompositions:
                # target can implement this op decomposed
                if hasattr(cmd.op, "decomp") and not cmd.op.decomp:
                    # user has requested application of the op as a primitive
                    if op_name in self.primitives:
                        compiled.append(cmd)
                        continue
                    else:
                        raise pu.CircuitError(
                            "The operation {} is not a primitive for the target '{}'".format(
                                cmd.op.__class__.__name__, self.short_name
                            )
                        )
                try:
                    kwargs = self.decompositions[op_name]
                    temp = cmd.op.decompose(cmd.reg, **kwargs)
                    # now compile the decomposition
                    temp = self.decompose(temp)
                    compiled.extend(temp)
                except NotImplementedError as err:
                    # Operation does not have _decompose() method defined!
                    # simplify the error message by suppressing the previous exception
                    raise err from None

            elif op_name in self.primitives:
                # target can handle the op natively
                compiled.append(cmd)

            else:
                raise pu.CircuitError(
                    "The operation {} cannot be used with the target '{}'.".format(
                        cmd.op.__class__.__name__, self.short_name
                    )
                )

        return compiled



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains various utility classes and functions used
within the :class:`~.Program` class.
"""

from collections.abc import Sequence

import networkx as nx

from .parameters import MeasuredParameter


__all__ = [
    "Program_current_context",
    "RegRefError",
    "CircuitError",
    "MergeFailure",
    "Command",
    "RegRef",
    "list_to_grid",
    "grid_to_DAG",
    "DAG_to_list",
    "list_to_DAG",
    "group_operations",
    "optimize_circuit",
]


Program_current_context = None
"""Context for inputting a Program. Used to be a class attribute of :class:`.Program`, placed
here to avoid cyclic imports."""
# todo: Avoid issues with Program contexts and threading,
# cf. _pydecimal.py in the python standard distribution.


class RegRefError(IndexError):
    """Exception raised by :class:`.Program` when it encounters an invalid register reference.

    E.g., trying to apply a gate to a nonexistent or deleted subsystem.
    """


class CircuitError(RuntimeError):
    """Exception raised by :class:`.Program` when it encounters an illegal
    operation in the quantum circuit.

    E.g., trying to use an Operation type that is unsupported by the current compilation target.
    """


class MergeFailure(RuntimeError):
    """Exception raised by :meth:`strawberryfields.ops.Operation.merge` when an
    attempted merge fails.

    E.g., trying to merge two gates of different families.
    """


class Command:
    """Represents a quantum operation applied on specific subsystems of the register.

    A Command instance is immutable once created, and can be shared between
    several :class:`.Program` instances.

    Args:
        op (~strawberryfields.ops.Operation): quantum operation to apply
        reg (Sequence[RegRef]): Subsystems to which the operation is applied.
            Note that the order matters here.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, op, reg):
        # accept a single RegRef in addition to a Sequence
        if not isinstance(reg, Sequence):
            reg = [reg]

        #: Operation: quantum operation to apply
        self.op = op
        #: Sequence[RegRef]: subsystems to which the operation is applied
        self.reg = reg

    def __str__(self):
        """
        Return a string containing the command in Blackbird syntax.
        """

        operation = str(self.op)
        if self.op.ns == 0:
            # op takes no subsystems as parameters, do not print anything more
            code = operation
        else:
            subsystems = ", ".join([str(r) for r in self.reg])
            code = "{} | ({})".format(operation, subsystems)
        return code

    def __lt__(self, other):
        # Needed as a tiebreaker for NetworkX lexicographical_topological_sort()
        # due to a buggy implementation! Any order will do. Remove when NetworkX is fixed.
        return True

    def get_dependencies(self):
        """Subsystems the command depends on.

        Combination of ``self.reg`` and ``self.op.measurement_deps``.

        .. note:: ``measurement_deps`` are used to ensure that the measurement
            happens before the result is used, but this is a bit too strict:
            two gates depending on the same measurement result but otherwise
            acting on different subsystems should commute.

        Returns:
            set[RegRef]: set of subsystems the command depends on
        """
        deps = self.op.measurement_deps | set(self.reg)
        return deps


class RegRef:
    """Quantum register reference.

    The objects of this class refer to a specific subsystem (mode) of
    a quantum register.

    Within the scope of each :class:`.Program` instance, only one RegRef instance
    should exist per subsystem. Program keeps the authoritative mapping
    of subsystem indices to RegRef instances.
    Subsystem measurement results are stored in the "official" RegRef object.
    If other RegRef objects referring to the same subsystem exist, they will
    not be updated. Once a RegRef is assigned a subsystem index it will never
    change, not even if the subsystem is deleted.

    The RegRefs are constructed in :meth:`.Program._add_subsystems`.

    Args:
        ind (int): index of the register subsystem referred to
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, ind):
        self.ind = ind  #: int: subsystem index
        self.val = None  #: float, complex: Measurement result. None if the subsystem has not been measured yet.
        self.active = True  #: bool: True at construction, False after the subsystem is deleted

    def __str__(self):
        return "q[{}]".format(self.ind)

    def __hash__(self):
        """Hashing method.

        NOTE: Has to match :meth:`__eq__` such that if two RegRefs compare equal they must have equal hashes.
        """
        return hash((self.ind, self.active))

    def __eq__(self, other):
        """Equality comparison.

        Compares the index and the activity state of the two RegRefs, the val field does not matter.
        NOTE: Affects the hashability of RegRefs, see also :meth:`__hash__`.
        """
        if other.__class__ != self.__class__:
            print("---------------          regref.__eq__: compared reqref to ", other.__class__)
            return False
        return self.ind == other.ind and self.active == other.active

    @property
    def par(self):
        """Convert the RegRef into a measured parameter.

        Returns:
            MeasuredParameter: measured parameter linked to this RegRef
        """
        return MeasuredParameter(self)


# =================
# Utility functions
# =================


def list_to_grid(ls):
    """Transforms a list of Commands to a grid representation.

    The grid is a mapping from subsystem indices to lists of :class:`Command` instances touching
    that subsystem, in temporal order. The same Command instance will appear in each list that
    corresponds to one of its subsystems.

    Args:
        ls (Iterable[Command]): quantum circuit
    Returns:
        dict[int, list[Command]]: same circuit in grid form
    """
    grid = {}
    # enter every operation in the list to its proper position in the grid
    for cmd in ls:
        for r in cmd.get_dependencies():
            # Add cmd to the grid to the end of the line r.ind.
            grid.setdefault(r.ind, []).append(cmd)
    return grid


def grid_to_DAG(grid):
    """Transforms a grid of Commands to a DAG representation.

    In the DAG (directed acyclic graph) each node is a :class:`Command` instance,
    and edges point from Commands to their immediate dependents/followers.

    Args:
        grid (dict[int, list[Command]]): quantum circuit
    Returns:
        networkx.DiGraph[Command]: same circuit in DAG form
    """
    DAG = nx.DiGraph()
    for _, q in grid.items():
        if q:
            # add the first operation on the wire that does not depend on anything
            DAG.add_node(q[0])
        for i in range(1, len(q)):
            # add the edge between the operations, and the operation nodes themselves
            DAG.add_edge(q[i - 1], q[i])
    return DAG


def list_to_DAG(ls):
    """Transforms a list of Commands to a DAG representation.

    In the DAG (directed acyclic graph) each node is a :class:`Command` instance,
    and edges point from Commands to their immediate dependents/followers.

    Args:
        ls (Iterable[Command]): quantum circuit
    Returns:
        networkx.DiGraph[Command]: same circuit in DAG form
    """
    return grid_to_DAG(list_to_grid(ls))


def DAG_to_list(dag):
    """Transforms a Command DAG to a list representation.

    The list contains the :class:`Command` instances in (one possible) topological order,
    i.e., dependants following the operations they depend on.

    Args:
        dag (networkx.DiGraph[Command]): quantum circuit
    Returns:
        list[Command]: same circuit in list form
    """
    # sort the operation graph into topological order
    temp = nx.algorithms.dag.topological_sort(dag)
    return list(temp)


def group_operations(seq, predicate):
    """Group a set of Operations in a circuit together (if possible).

    For the purposes of this method, we call a :class:`Operation` instance *marked* iff
    ``predicate`` returns True on it.

    This method converts the quantum circuit in ``seq`` into an equivalent circuit ``A+B+C``,
    where the :class:`Command` instances in sequences ``A`` and ``C`` do not contain any
    marked Operations.
    The sequence ``B`` contains all marked Operations in the circuit, and possibly
    additional unmarked instances that could not be moved into ``A`` or ``C`` using the
    available commutation rules.
    Any of the three returned sequences can be empty (but if ``B`` is empty then so is ``C``).

    Args:
        seq (Sequence[Command]): quantum circuit
        predicate (Callable[[Operation], bool]): Grouping predicate. Returns True for the
            Operations to be grouped together, False for the others.
    Returns:
        Tuple[Sequence[Command]]: A, B, C such that A+B+C is equivalent to seq,
            and A and C do not contain any marked Operation instances.
    """

    def find_first_index(seq):
        """Index of the first element in the sequence for which the predicate function returns True.
        If no such element exists, returns the length of the sequence.
        """
        return next((i for i, e in enumerate(seq) if predicate(e.op)), len(seq))

    def marked_last(node):
        """Mapping from nodes to sorting keys to resolve ambiguities in the topological sort.
        Larger key values come later in the lexicographical-topological ordering.
        """
        if predicate(node.op):
            return 1
        return 0

    def lex_topo(seq, key):
        """Sorts a Command sequence lexicographical-topologically using the given lexicographic key function."""
        DAG = list_to_DAG(seq)
        return list(nx.algorithms.dag.lexicographical_topological_sort(DAG, key=key))

    C = lex_topo(seq, key=marked_last)
    ind = find_first_index(C)
    A = C[:ind]  # initial unmarked instances
    B = C[ind:]  # marked and possibly unmarked

    # re-sort B, marked instances first
    C = lex_topo(B, key=lambda x: -marked_last(x))
    # find last marked
    ind = len(C) - find_first_index(list(reversed(C)))
    B = C[:ind]  # marked and still possibly unmarked
    C = C[ind:]  # final unmarked instances
    return A, B, C


def optimize_circuit(seq):
    """Try to simplify and optimize a quantum circuit.

    The purpose of the optimizer is to simplify the circuit
    to make it cheaper and faster to execute. Different backends may require
    different types of optimization, but in general the fewer operations a circuit has,
    the faster it should run. The optimizer thus should convert the circuit into a
    simpler :term:`equivalent circuit`.

    The optimizations are based on the abstract algebraic properties of the Operations
    constituting the circuit, e.g., combining two consecutive gates of the same gate family,
    and at no point should require a matrix representation of any kind.
    The optimization must also not change the state of the RegRefs in any way.

    Currently the optimization is very simple. It

    * merges neighboring :class:`state preparations <.Preparation>` and :class:`gates <.Gate>`
      belonging to the same family and acting on the same sequence of subsystems
    * cancels neighboring pairs of a gate and its inverse

    Args:
        seq (Sequence[Command]): quantum circuit to optimize

    Returns:
        List[Command]: optimized circuit
    """

    def _print_list(i, q, print_fn=print):
        "For debugging."
        # pylint: disable=unreachable
        return
        print_fn("i: {},  len: {}   ".format(i, len(q)), end="")
        for x in q:
            print_fn(x.op, ", ", end="")
        print_fn()

    grid = list_to_grid(seq)

    # try merging neighboring operations on each wire
    # TODO the merging could also be done using the circuit DAG, which
    # might be smarter (ns>1 would be easy)
    for k in grid:
        q = grid[k]
        i = 0  # index along the wire
        _print_list(i, q)
        while i + 1 < len(q):
            # at least two operations left to merge on this wire
            try:
                a = q[i]
                b = q[i + 1]
                # the ops must have equal size and act on the same wires
                if a.op.ns == b.op.ns and a.reg == b.reg:
                    if a.op.ns != 1:
                        # ns > 1 is tougher. on no wire must there be anything
                        # between them, also deleting is more complicated
                        # todo treat it as a failed merge for now
                        i += 1
                        continue
                    op = a.op.merge(b.op)
                    # merge was successful, delete the old ops
                    del q[i : i + 2]
                    # insert the merged op (unless it's identity)
                    if op is not None:
                        q.insert(i, Command(op, a.reg))
                    # move one spot backwards to try another merge
                    if i > 0:
                        i -= 1
                    _print_list(i, q)
                    continue
            except MergeFailure:
                pass
            i += 1  # failed at merging the ops, move forward

    # convert the circuit back into a list (via a DAG)
    DAG = grid_to_DAG(grid)
    return DAG_to_list(DAG)



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing Gaussian backend specific extensions to BaseGaussianState"""

import numpy as np

from ..states import BaseGaussianState
from .ops import fock_amplitudes_one_mode, fock_prob, sm_fidelity


class GaussianState(BaseGaussianState):
    r"""Class for the representation of quantum states using Gaussian representation.

    Args:
        state_data (tuple(mu, cov)): A tuple containing the vector of means ``mu`` and the
            covariance matrix ``cov``.
        num_modes (int): the number of modes in the state
        qmat (array): The covariance matrix for the Q function
        Amat (array): The A matrix from Hamilton's paper
        mode_names (Sequence): (optional) this argument contains a list providing mode names
            for each mode in the state
    """

    def __init__(self, state_data, num_modes, qmat, Amat, mode_names=None):
        # pylint: disable=too-many-arguments
        super().__init__(state_data, num_modes, mode_names)
        self.qmat = qmat
        self.Amat = Amat

    def reduced_dm(self, modes, **kwargs):
        r"""Returns the reduced density matrix in the Fock basis for a particular mode.

        Args:
            modes (int): specifies the mode. For the Gaussian backend, only a **single** mode
                reduced density matrix can be returned.
            **kwargs:

                  * **cutoff** (*int*): (default 10) specifies where to truncate the returned density matrix.
                    Note that the cutoff argument only applies for Gaussian representation;
                    states represented in the Fock basis will use their own internal cutoff dimension.

        Returns:
            array: the reduced density matrix for the specified modes
        """
        cutoff = kwargs.get("cutoff", 10)
        mu, cov = self.reduced_gaussian([modes])  # pylint: disable=unused-variable
        cov = cov * 2 / self._hbar

        return fock_amplitudes_one_mode(self._alpha[modes], cov, cutoff)

    # ==========================================
    # The methods below inherit their docstring
    # from BaseState

    def fock_prob(self, n, **kwargs):
        if len(n) != self._modes:
            raise ValueError("Fock state must be same length as the number of modes")

        cutoff = kwargs.get("cutoff", 10)
        if sum(n) >= cutoff:
            raise ValueError("Cutoff argument must be larger than the sum of photon numbers")

        ocp = np.uint8(np.array(n))

        # get the vector and means and covariance
        # matrix, and normalize so that hbar=2
        mu = self.means() * np.sqrt(2 / self.hbar)
        cov = self.cov() * 2 / self.hbar

        return fock_prob(mu, cov, ocp)

    def mean_photon(self, mode, **kwargs):
        mu, cov = self.reduced_gaussian([mode])
        mean = (np.trace(cov) + mu.T @ mu) / (2 * self._hbar) - 1 / 2
        var = (np.trace(cov @ cov) + 2 * mu.T @ cov @ mu) / (2 * self._hbar ** 2) - 1 / 4
        return mean, var

    def fidelity(self, other_state, mode, **kwargs):
        mu1 = other_state[0] * 2 / np.sqrt(2 * self._hbar)
        cov1 = other_state[1] * 2 / self._hbar

        mu2, cov2 = self.reduced_gaussian([mode])
        mu2 *= 2 / np.sqrt(2 * self._hbar)
        cov2 /= self._hbar / 2

        return sm_fidelity(mu1, mu2, cov1, cov2)

    def fidelity_vacuum(self, **kwargs):
        alpha = np.zeros(len(self._alpha))
        return self.fidelity_coherent(alpha)

    def fidelity_coherent(self, alpha_list, **kwargs):
        if len(alpha_list) != self._modes:
            raise ValueError("alpha_list must be same length as the number of modes")

        Qi = np.linalg.inv(self.qmat)
        delta = self._alpha - alpha_list

        delta = np.concatenate((delta, delta.conj()))
        fac = np.sqrt(np.linalg.det(Qi).real)
        exp = np.exp(-0.5 * np.dot(delta, np.dot(Qi, delta.conj())).real)
        return fac * exp



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Tools for users to find dense subgraphs of different sizes.

The search heuristic in this module works by resizing a collection of starting subgraphs and
keeping track of the densest identified. The starting subgraphs can be selected by sampling from
GBS, resulting in candidates that are likely to be dense :cite:`arrazola2018using`.

.. seealso::

    :ref:`apps-subgraph-tutorial`

Algorithm
^^^^^^^^^

The heuristic algorithm provided proceeds as follows. Each starting subgraph :math:`s` is resized
to a range of sizes :math:`\{k\}_{k_{\min}}^{k_{\max}}`, resulting in the subgraphs
:math:`s_{k}`. For a given size :math:`k`, a collection of the densest :math:`n` subgraphs
identified is recorded, meaning that :math:`s_{k}` is added to the collection only if it has
sufficient density.

This algorithm returns a dictionary over the range of sizes specified, with each value being the
collection of densest-:math:`k` subgraphs. This collection is a list of tuple pairs specifying
the subgraph density and nodes.

Subgraph resizing
^^^^^^^^^^^^^^^^^

The key element of the :func:`search` algorithm is the resizing of each subgraph, allowing a
range of subgraph sizes to be tracked. Resizing proceeds in the :func:`resize` function by greedily
adding or removing nodes to a subgraph one-at-a-time. Node selection is carried out by picking
the node with the greatest or least degree with respect to the subgraph. This function returns a
dictionary over the range of sizes specified, with each value being the corresponding resized
subgraph.

Whenever there are multiple candidate nodes with the same degree, there must be a choice of
which node to add or remove. The supported choices are:

- Select among candidate nodes uniformly at random;
- Select the candidate node with the greatest node weight, settling remaining ties uniformly at
  random.
"""
from typing import Tuple, Union

import networkx as nx
import numpy as np


def search(
    subgraphs: list,
    graph: nx.Graph,
    min_size: int,
    max_size: int,
    max_count: int = 10,
    node_select: Union[str, np.ndarray, list] = "uniform",
) -> dict:
    """Search for dense subgraphs within an input size range.

    For each subgraph from ``subgraphs``, this function resizes using :func:`resize` to the input
    range specified by ``min_size`` and ``max_size``, resulting in a range of differently sized
    subgraphs. This function loops over all elements of ``subgraphs`` and keeps track of the
    ``max_count`` number of densest subgraphs identified for each size.

    In both growth and shrink phases of :func:`resize`, there may be multiple candidate nodes with
    equal degree to add to or remove from the subgraph. The method of selecting the node is
    specified by the ``node_select`` argument, which can be either:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the highest weight (when growing) and lowest weight (when shrinking),
      settling remaining ties by uniform random choice.

    **Example usage:**

    >>> s = data.Planted()
    >>> g = nx.Graph(s.adj)
    >>> s = sample.postselect(s, 16, 30)
    >>> s = sample.to_subgraphs(s, g)
    >>> search(s, g, 8, 9, max_count=3)
    {9: [(0.9722222222222222, [21, 22, 23, 24, 25, 26, 27, 28, 29]),
      (0.9722222222222222, [20, 21, 22, 24, 25, 26, 27, 28, 29]),
      (0.9444444444444444, [20, 21, 22, 23, 24, 25, 26, 27, 29])],
     8: [(1.0, [21, 22, 24, 25, 26, 27, 28, 29]),
      (1.0, [21, 22, 23, 24, 25, 26, 27, 28]),
      (1.0, [20, 21, 22, 24, 25, 26, 27, 29])]}

    Args:
        subgraphs (list[list[int]]): a list of subgraphs specified by their nodes
        graph (nx.Graph): the input graph
        min_size (int): minimum size to search for dense subgraphs
        max_size (int): maximum size to search for dense subgraphs
        max_count (int): maximum number of densest subgraphs to keep track of for each size
        node_select (str, list or array): method of settling ties when more than one node of
            equal degree can be added/removed. Can be ``"uniform"`` (default), or a NumPy array or
            list containing node weights.

    Returns:
        dict[int, list[tuple[float, list[int]]]]: a dictionary of different sizes, each containing a
        list of densest subgraphs reported as a tuple of subgraph density and subgraph nodes,
        sorted in non-increasing order of density
    """
    dense = {}

    for s in subgraphs:
        r = resize(s, graph, min_size, max_size, node_select)

        for size, subgraph in r.items():
            r[size] = (nx.density(graph.subgraph(subgraph)), subgraph)

        _update_dict(dense, r, max_count)

    return dense


def _update_dict(d: dict, d_new: dict, max_count: int) -> None:
    """Updates dictionary ``d`` with subgraph tuples contained in ``d_new``.

    Subgraph tuples are a pair of values: a float specifying the subgraph density and a list of
    integers specifying the subgraph nodes. Both ``d`` and ``d_new`` are dictionaries over
    different subgraph sizes. The values of ``d`` are lists of subgraph tuples containing the top
    densest subgraphs for a given size, with maximum length ``max_count``. The values of
    ``d_new`` are candidate subgraph tuples that can be the result of resizing an input subgraph
    over a range using :func:`resize`. We want to add these candidates to the list of subgraph
    tuples in ``d`` to build up our collection of dense subgraphs.

    Args:
        d (dict[int, list[tuple[float, list[int]]]]): dictionary of subgraph sizes and
            corresponding list of subgraph tuples
        d_new (dict[int, tuple[float, list[int]]]): dictionary of subgraph sizes and corresponding
            subgraph tuples that are candidates to be added to the list
        max_count (int):  the maximum length of every subgraph tuple list

    Returns:
        None: this function modifies the dictionary ``d`` in place
    """
    for size, t in d_new.items():
        l = d.setdefault(size, [t])
        _update_subgraphs_list(l, t, max_count)


def _update_subgraphs_list(l: list, t: tuple, max_count: int) -> None:
    """Updates list of top subgraphs with a candidate.

    Here, the list ``l`` to be updated is a list of tuples with each tuple being a pair of
    values: a float specifying the subgraph density and a list of integers specifying the
    subgraph nodes. For example, ``l`` may be:

    ``[(0.8, [0, 5, 9, 10]), (0.5, [1, 2, 5, 6]), (0.3, [0, 4, 6, 9])]``

    We want to update ``l`` with a candidate tuple ``t``, which should be a pair specifying a
    subgraph density and corresponding subgraph nodes. For example, we might want to add:

    ``(0.4, [1, 4, 9, 10])``

    This function checks:

    - if ``t`` is already an element of ``l``, do nothing (i.e., so that ``l`` never has
      repetitions)

    - if ``len(l) < max_count``, add ``t``

    - otherwise, if the density of ``t`` exceeds the minimum density of ``l`` , add ``t`` and
      remove the element with the minimum density

    - otherwise, if the density of ``t`` equals the minimum density of ``l``, flip a coin and
      randomly swap in ``t`` with the minimum element of ``l``.

    The list ``l`` is also sorted so that its first element is the subgraph with the highest
    density.

    Args:
        l (list[tuple[float, list[int]]]): the list of subgraph tuples to be updated
        t (tuple[float, list[int]): the candidate subgraph tuple
        max_count (int): the maximum length of ``l``

    Returns:
        None: this function modifies ``l`` in place
    """
    t = (t[0], sorted(set(t[1])))

    for _d, s in l:
        if t[1] == s:
            return

    if len(l) < max_count:
        l.append(t)
        l.sort(reverse=True)
        return

    l_min = l[-1][0]

    if t[0] > l_min:
        l.append(t)
        l.sort(reverse=True)
        del l[-1]
    elif t[0] == l_min:
        if np.random.choice(2):
            del l[-1]
            l.append(t)
            l.sort(reverse=True)

    return


def resize(
    subgraph: list,
    graph: nx.Graph,
    min_size: int,
    max_size: int,
    node_select: Union[str, np.ndarray, list] = "uniform",
) -> dict:
    """Resize a subgraph to a range of input sizes.

    This function uses a greedy approach to iteratively add or remove nodes one at a time to an
    input subgraph to reach the range of sizes specified by ``min_size`` and ``max_size``.

    When growth is required, the algorithm examines all nodes from the remainder of the graph as
    candidates and adds the single node with the highest degree relative to the rest of the
    subgraph. This results in a graph that is one node larger, and if growth is still required,
    the algorithm performs the procedure again.

    When shrinking is required, the algorithm examines all nodes from within the subgraph as
    candidates and removes the single node with lowest degree relative to the subgraph.

    In both growth and shrink phases, there may be multiple candidate nodes with equal degree to
    add to or remove from the subgraph. The method of selecting the node is specified by the
    ``node_select`` argument, which can be either:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the highest weight (when growing) and lowest weight (when shrinking),
      settling remaining ties by uniform random choice.

    **Example usage:**

    >>> s = data.Planted()
    >>> g = nx.Graph(s.adj)
    >>> s = [20, 21, 22, 23, 24, 25, 26, 27, 28, 29]
    >>> resize(s, g, 8, 12)
    {10: [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
     11: [11, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
     12: [0, 11, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
     9: [20, 21, 22, 24, 25, 26, 27, 28, 29],
     8: [20, 21, 22, 24, 25, 26, 27, 29]}

    Args:
        subgraph (list[int]): a subgraph specified by a list of nodes
        graph (nx.Graph): the input graph
        min_size (int): minimum size for subgraph to be resized to
        max_size (int): maximum size for subgraph to be resized to
        node_select (str, list or array): method of settling ties when more than one node of
            equal degree can be added/removed. Can be ``"uniform"`` (default), or a NumPy array or
            list containing node weights.

    Returns:
        dict[int, list[int]]: a dictionary of different sizes with corresponding subgraph
    """
    nodes = graph.nodes()
    subgraph = set(subgraph)
    node_select, w = _validate_inputs(subgraph, graph, min_size, max_size, node_select)

    starting_size = len(subgraph)

    if min_size <= starting_size <= max_size:
        resized = {starting_size: sorted(subgraph)}
    else:
        resized = {}

    if max_size > starting_size:

        grow_subgraph = graph.subgraph(subgraph).copy()

        while grow_subgraph.order() < max_size:
            grow_nodes = grow_subgraph.nodes()
            complement_nodes = nodes - grow_nodes

            degrees = np.array(
                [(c, graph.subgraph(list(grow_nodes) + [c]).degree()[c]) for c in complement_nodes]
            )
            degrees_max = np.argwhere(degrees[:, 1] == degrees[:, 1].max()).flatten()

            if node_select == "uniform":
                to_add_index = np.random.choice(degrees_max)
            elif node_select == "weight":
                weights = np.array([w[degrees[n][0]] for n in degrees_max])
                to_add_index = np.random.choice(np.where(weights == weights.max())[0])

            to_add = degrees[to_add_index][0]
            grow_subgraph.add_node(to_add)
            new_size = grow_subgraph.order()

            if min_size <= new_size <= max_size:
                resized[new_size] = sorted(grow_subgraph.nodes())

    if min_size < starting_size:

        shrink_subgraph = graph.subgraph(subgraph).copy()

        while shrink_subgraph.order() > min_size:
            degrees = np.array(shrink_subgraph.degree)
            degrees_min = np.argwhere(degrees[:, 1] == degrees[:, 1].min()).flatten()

            if node_select == "uniform":
                to_remove_index = np.random.choice(degrees_min)
            elif node_select == "weight":
                weights = np.array([w[degrees[n][0]] for n in degrees_min])
                to_remove_index = np.random.choice(np.where(weights == weights.min())[0])

            to_remove = degrees[to_remove_index][0]
            shrink_subgraph.remove_node(to_remove)

            new_size = shrink_subgraph.order()

            if min_size <= new_size <= max_size:
                resized[new_size] = sorted(shrink_subgraph.nodes())

    return resized


def _validate_inputs(
    subgraph: set,
    graph: nx.Graph,
    min_size: int,
    max_size: int,
    node_select: Union[str, np.ndarray, list] = "uniform",
) -> Tuple:
    """Validates input for the ``resize`` function.

    This function checks:
        - if ``subgraph`` is a valid subgraph of ``graph``;
        - if ``min_size`` and ``max_size`` are sensible numbers;
        - if ``node_select`` is either ``"uniform"`` or a NumPy array or list;
        - if, when ``node_select`` is a NumPy array or list, that it is the correct size and that
          ``node_select`` is changed to ``"weight"``.

    This function returns the updated ``node_select`` and a dictionary mapping nodes to their
    corresponding weights (weights default to unity if not specified).

    Args:
        subgraph (list[int]): a subgraph specified by a list of nodes
        graph (nx.Graph): the input graph
        min_size (int): minimum size for subgraph to be resized to
        max_size (int): maximum size for subgraph to be resized to
        node_select (str, list or array): method of settling ties when more than one node of
            equal degree can be added/removed. Can be ``"uniform"`` (default), or a NumPy array or
            list containing node weights.

    Returns:
        tuple[str, dict]: the updated ``node_select`` and a dictionary of node weights
    """
    if not subgraph.issubset(graph.nodes()):
        raise ValueError("Input is not a valid subgraph")
    if min_size < 1:
        raise ValueError("min_size must be at least 1")
    if max_size >= len(graph.nodes()):
        raise ValueError("max_size must be less than number of nodes in graph")
    if max_size < min_size:
        raise ValueError("max_size must not be less than min_size")

    if isinstance(node_select, (list, np.ndarray)):
        if len(node_select) != graph.number_of_nodes():
            raise ValueError("Number of node weights must match number of nodes")
        w = {n: node_select[i] for i, n in enumerate(graph.nodes)}
        node_select = "weight"
    else:
        w = {n: 1 for i, n in enumerate(graph.nodes)}
        if node_select != "uniform":
            raise ValueError("Node selection method not recognized")

    return node_select, w



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package contains the modules that make up the
Strawberry Fields application layer.

.. currentmodule:: strawberryfields.apps
.. autosummary::
    :toctree: api

	clique
	data
	plot
	points
	sample
	similarity
	subgraph
	vibronic
"""

import strawberryfields.apps.clique
import strawberryfields.apps.data
import strawberryfields.apps.plot
import strawberryfields.apps.points
import strawberryfields.apps.sample
import strawberryfields.apps.similarity
import strawberryfields.apps.subgraph
import strawberryfields.apps.vibronic



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit specifications for the Gaussian simulator backend."""

import numpy as np
from strawberryfields.program_utils import Command
from strawberryfields import ops
from strawberryfields.parameters import par_evaluate
from thewalrus.symplectic import (
    expand_vector,
    expand,
    rotation,
    squeezing,
    two_mode_squeezing,
    interferometer,
    beam_splitter,
)
from .circuit_specs import CircuitSpecs


class GaussianUnitary(CircuitSpecs):
    """Compiler to arrange a Gaussian quantum circuit into the canonical Symplectic form.

    This compile specification checks whether the circuit can be implemented as a sequence of
    Gaussian operations. If so, it arranges them in the canonical order with displacement at the end.
    After compilation, the circuit will consist of at most two operations, a :class:`~.GaussianTransform`
    and a :class:`~.Dgate`.

    This compiler can be accessed by calling :meth:`.Program.compile` with `'gaussian_unitary'` specified.

    **Example:**

    Consider the following Strawberry Fields program, compiled using the `'gaussian_unitary'` compiler:

    .. code-block:: python3

        from strawberryfields.ops import Xgate, Zgate, Sgate, Dgate, Rgate
        import strawberryfields as sf

        circuit = sf.Program(1)
        with circuit.context as q:
            Xgate(0.4) | q[0]
            Zgate(0.5) | q[0]
            Sgate(0.6) | q[0]
            Dgate(1.0+2.0j) | q[0]
            Rgate(0.3) | q[0]
            Sgate(0.6, 1.0) | q[0]

        compiled_circuit = circuit.compile("gaussian_unitary")

    We can now print the compiled circuit, consisting of one
    :class:`~.GaussianTransform` and one :class:`~.Dgate`:

    >>> compiled_circuit.print()
    GaussianTransform([[ 0.3543 -1.3857]
                       [-0.0328  2.9508]]) | (q[0])
    Dgate(-1.151+3.91j, 0) | (q[0])
    """

    short_name = "gaussian_unitary"
    modes = None
    local = True
    remote = True
    interactive = True

    primitives = {
        # meta operations
        "All",
        "_New_modes",
        "_Delete",
        # single mode gates
        "Dgate",
        "Sgate",
        "Rgate",
        # multi mode gates
        "BSgate",
        "S2gate",
        "Interferometer",  # Note that interferometer is accepted as a primitive
        "GaussianTransform",  # Note that GaussianTransform is accepted as a primitive
    }

    decompositions = {
        "GraphEmbed": {},
        "BipartiteGraphEmbed": {},
        "Gaussian": {},
        "Pgate": {},
        "CXgate": {},
        "CZgate": {},
        "MZgate": {},
        "Xgate": {},
        "Zgate": {},
        "Fouriergate": {},
    }
    # pylint: disable=too-many-branches
    def compile(self, seq, registers):
        """Try to arrange a quantum circuit into the canonical Symplectic form.

        This method checks whether the circuit can be implemented as a sequence of Gaussian operations.
        If the answer is yes it arranges them in the canonical order with displacement at the end.

        Args:
            seq (Sequence[Command]): quantum circuit to modify
            registers (Sequence[RegRefs]): quantum registers
        Returns:
            List[Command]: modified circuit
        Raises:
            CircuitError: the circuit does not correspond to a Gaussian unitary
        """

        # Check which modes are actually being used
        used_modes = []
        for operations in seq:
            modes = [modes_label.ind for modes_label in operations.reg]
            used_modes.append(modes)
        # pylint: disable=consider-using-set-comprehension
        used_modes = list(set([item for sublist in used_modes for item in sublist]))

        # dictionary mapping the used modes to consecutive non-negative integers
        dict_indices = {used_modes[i]: i for i in range(len(used_modes))}
        nmodes = len(used_modes)

        # This is the identity transformation in phase-space, multiply by the identity and add zero
        Snet = np.identity(2 * nmodes)
        rnet = np.zeros(2 * nmodes)

        # Now we will go through each operation in the sequence `seq` and apply it in quadrature space
        # We will keep track of the net transforation in the Symplectic matrix `Snet` and the quadrature
        # vector `rnet`.
        for operations in seq:
            name = operations.op.__class__.__name__
            params = par_evaluate(operations.op.p)
            modes = [modes_label.ind for modes_label in operations.reg]
            if name == "Dgate":
                rnet = rnet + expand_vector(
                    params[0] * (np.exp(1j * params[1])), dict_indices[modes[0]], nmodes
                )
            else:
                if name == "Rgate":
                    S = expand(rotation(params[0]), dict_indices[modes[0]], nmodes)
                elif name == "Sgate":
                    S = expand(squeezing(params[0], params[1]), dict_indices[modes[0]], nmodes)
                elif name == "S2gate":
                    S = expand(
                        two_mode_squeezing(params[0], params[1]),
                        [dict_indices[modes[0]], dict_indices[modes[1]]],
                        nmodes,
                    )
                elif name == "Interferometer":
                    S = expand(
                        interferometer(params[0]), [dict_indices[mode] for mode in modes], nmodes
                    )
                elif name == "GaussianTransform":
                    S = expand(params[0], [dict_indices[mode] for mode in modes], nmodes)
                elif name == "BSgate":
                    S = expand(
                        beam_splitter(params[0], params[1]),
                        [dict_indices[modes[0]], dict_indices[modes[1]]],
                        nmodes,
                    )
                Snet = S @ Snet
                rnet = S @ rnet

        # Having obtained the net displacement we simply convert it into complex notation
        alphas = 0.5 * (rnet[0:nmodes] + 1j * rnet[nmodes : 2 * nmodes])
        # And now we just pass the net transformation as a big Symplectic operation plus displacements
        ord_reg = [r for r in list(registers) if r.ind in used_modes]
        ord_reg = sorted(list(ord_reg), key=lambda x: x.ind)
        if np.allclose(Snet, np.identity(2 * nmodes)):
            A = []
        else:
            A = [Command(ops.GaussianTransform(Snet), ord_reg)]
        B = [
            Command(ops.Dgate(alphas[i]), ord_reg[i])
            for i in range(len(ord_reg))
            if not np.allclose(alphas[i], 0.0)
        ]
        return A + B



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The Strawberry Fields codebase includes a number of complementary components.
These can be separated into frontend components, applications layer,
and backend components (all found within the :mod:`strawberryfields.backends` submodule).

.. image:: ../_static/sfcomponents.svg
    :align: center
    :width: 90%
    :target: javascript:void(0);
"""
from . import apps
from ._version import __version__
from .cli import ping
from .configuration import store_account
from .engine import Engine, LocalEngine, RemoteEngine
from .io import load, save
from .parameters import par_funcs as math
from .program import Program

__all__ = ["Engine", "RemoteEngine", "Program", "version", "save", "load", "about", "cite"]


#: float: numerical value of hbar for the frontend (in the implicit units of position * momentum)
hbar = 2


def version():
    r"""
    Version number of Strawberry Fields.

    Returns:
      str: package version number
    """
    return __version__


def about():
    """Strawberry Fields information.

    Prints the installed version numbers for SF and its dependencies,
    and some system info. Please include this information in bug reports.

    **Example:**

    .. code-block:: pycon

        >>> sf.about()
        Strawberry Fields: a Python library for continuous-variable quantum circuits.
        Copyright 2018-2020 Xanadu Quantum Technologies Inc.

        Python version:            3.6.8
        Platform info:             Linux-5.0.0-36-generic-x86_64-with-debian-buster-sid
        Installation path:         /home/josh/Dropbox/Work/Xanadu/sf_cloud/strawberryfields
        Strawberry Fields version: 0.12.0-dev
        Numpy version:             1.17.4
        Scipy version:             1.3.0
        Sympy version:             1.5
        NetworkX version:          2.4
        The Walrus version:        0.10.0
        Blackbird version:         0.2.1
        TensorFlow version:        2.0.0
    """
    # pylint: disable=import-outside-toplevel
    import sys
    import platform
    import os
    import numpy
    import scipy
    import sympy
    import networkx
    import thewalrus
    import blackbird

    # a QuTiP-style infobox
    print("\nStrawberry Fields: a Python library for continuous-variable quantum circuits.")
    print("Copyright 2018-2020 Xanadu Quantum Technologies Inc.\n")

    print("Python version:            {}.{}.{}".format(*sys.version_info[0:3]))
    print("Platform info:             {}".format(platform.platform()))
    print("Installation path:         {}".format(os.path.dirname(__file__)))
    print("Strawberry Fields version: {}".format(__version__))
    print("Numpy version:             {}".format(numpy.__version__))
    print("Scipy version:             {}".format(scipy.__version__))
    print("SymPy version:             {}".format(sympy.__version__))
    print("NetworkX version:          {}".format(networkx.__version__))
    print("The Walrus version:        {}".format(thewalrus.__version__))
    print("Blackbird version:         {}".format(blackbird.__version__))

    try:
        import tensorflow

        tf_version = tensorflow.__version__
    except ImportError:
        tf_version = None

    print("TensorFlow version:        {}".format(tf_version))


def cite():
    """Prints the BibTeX citation for Strawberry Fields.

    **Example:**

    .. code-block:: pycon

        >>> sf.cite()
        @article{strawberryfields,
            title = {{S}trawberry {F}ields: A Software Platform for Photonic Quantum Computing},
            author = {Killoran, Nathan and Izaac, Josh and Quesada, Nicol{'{a}}s and Bergholm, Ville and Amy, Matthew and Weedbrook, Christian},
            journal = {Quantum},
            volume = {3},
            pages = {129},
            year = {2019},
            doi = {10.22331/q-2019-03-11-129},
            archivePrefix = {arXiv},
            eprint = {1804.03159},
        }
    """
    citation = """@article{strawberryfields,
    title = {{S}trawberry {F}ields: A Software Platform for Photonic Quantum Computing},
    author = {Killoran, Nathan and Izaac, Josh and Quesada, Nicol{\'{a}}s and Bergholm, Ville and Amy, Matthew and Weedbrook, Christian},
    journal = {Quantum},
    volume = {3},
    pages = {129},
    year = {2019},
    doi = {10.22331/q-2019-03-11-129},
    archivePrefix = {arXiv},
    eprint = {1804.03159},
}"""
    print(citation)



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Pre-calculated datasets of simulated GBS samples.

.. seealso::

    :doc:`/introduction/data`
"""
# pylint: disable=unnecessary-pass
from abc import ABC, abstractmethod

import pkg_resources
import numpy as np
import scipy

DATA_PATH = pkg_resources.resource_filename("strawberryfields", "apps/data") + "/"


class Dataset(ABC):
    """Base class for loading datasets of pre-generated samples.

    Attributes:
        n_mean (float): mean number of photons in the GBS device
        threshold (bool): flag to indicate whether samples are generated with threshold detection
            (i.e., detectors of zero or some photons) or with photon-number-resolving detectors.
        n_samples (int): total number of samples in the dataset
        modes (int): number of modes in the GBS device or, equivalently, number of nodes in graph
        data (sparse): raw data of samples from GBS as a `csr sparse array
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.html>`__.
    """

    _count = 0

    @property
    @abstractmethod
    def _data_filename(self) -> str:
        """Base name of files containing the sample data stored in the ``./data/`` directory.

        Samples and corresponding adjacency matrix should both be provided as a
        ``scipy.sparse.csr_matrix`` saved in ``.npz`` format.

        For ``_data_filename = "example"``, the corresponding samples should be stored as
        ``./data/example.npz`` and the adjacency matrix as ``./data/example_A.npz``."""
        pass

    def __init__(self):
        self.data = scipy.sparse.load_npz(DATA_PATH + self._data_filename + ".npz")
        self.n_samples, self.modes = self.data.shape

    def __iter__(self):
        return self

    def __next__(self):
        if self._count < self.n_samples:
            self._count += 1
            return self.__getitem__(self._count - 1)
        self._count = 0
        raise StopIteration

    def _elem(self, i):
        """Access the i-th element of the sparse array and output as a list."""
        return list(self.data[i].toarray()[0])

    def __getitem__(self, key):

        if not isinstance(key, (slice, tuple, int)):
            raise TypeError("Dataset indices must be integers, slices, or tuples")

        if isinstance(key, int):
            return self._elem(key + self.n_samples if key < 0 else key)

        if isinstance(key, tuple):
            key = slice(*key)

        range_tuple = key.indices(self.n_samples)
        return [self._elem(i) for i in range(*range_tuple)]

    def __len__(self):
        return self.n_samples

    def counts(self, axis: int = 1) -> list:
        """Count number of photons or clicks.

        Counts number of photons/clicks in each sample (``axis==1``) or number of photons/clicks
        in each mode compounded over all samples (``axis==0``).

        Args:
            axis (int): axis to perform count

        Returns:
            list: counts from samples
        """
        return np.array(self.data.sum(axis)).flatten().tolist()

    # pylint: disable=missing-docstring
    @property
    @abstractmethod
    def n_mean(self) -> float:
        pass

    # pylint: disable=missing-docstring
    @property
    @abstractmethod
    def threshold(self) -> bool:
        pass


# pylint: disable=abstract-method
class GraphDataset(Dataset, ABC):
    """Class for loading datasets of pre-generated samples from graphs.

    Attributes:
        adj (array): adjacency matrix of the graph from which samples were generated
    """

    def __init__(self):
        super().__init__()
        self.adj = scipy.sparse.load_npz(DATA_PATH + self._data_filename + "_A.npz").toarray()


class Planted(GraphDataset):
    """A random 30-node graph containing a dense 10-node subgraph planted inside
    :cite:`arrazola2018using`.

    The graph is generated by joining two Erdős–Rényi random graphs. The first 20-node graph is
    generated with edge probability of 0.5 and the second 10-node planted graph is generated with
    edge probability of 0.875. The two graphs are joined by selecting 8 vertices at random from
    both and adding an edge between them.

    The 10-node planted clique is contained within the final 10 nodes of the graph.

    **Graph:**

    .. |planted| image:: ../../_static/graphs/planted.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |planted|

    Attributes:
        n_mean = 8
        threshold = True
        n_samples = 50000
        modes = 30
    """

    _data_filename = "planted"
    n_mean = 8
    threshold = True


class TaceAs(GraphDataset):
    """Binding interaction graph for the TACE-AS complex :cite:`banchi2019molecular`.

    Nodes in this graph correspond to pairs of atoms in a target protein and a pharmaceutical
    molecule. Edges in the graph are added if the distance between both pairs of atoms is very
    close to equal. Cliques in the graph correspond to possible docking configurations of protein
    and molecule, and the largest clique is the most stable configuration. There are multiple
    maximum-sized cliques of 8 nodes in this graph.

    **Graph:**

    .. |tace_as| image:: ../../_static/graphs/TACE-AS.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |tace_as|

    Attributes:
        n_mean = 8
        threshold = True
        n_samples = 50000
        modes = 24
    """

    _data_filename = "TACE-AS"
    n_mean = 8
    threshold = True


class PHat(GraphDataset):
    """Random graph created using the p-hat generator of :cite:`gendreau1993solving`.

    This graph is the ``p_hat300-1`` graph of the `DIMACS
    <http://iridia.ulb.ac.be/~fmascia/maximum_clique/DIMACS-benchmark>`__ dataset, which is a
    collection of large graphs with cliques that are hard to find. The best known clique of
    this 300-node graph is of size 8 and is composed of nodes: ``[53, 123, 180, 218, 246, 267, 270,
    286]``. This graph is not visualized due to its large size.

    Attributes:
        n_mean = 10
        threshold = True
        n_samples = 50000
        modes = 300
    """

    _data_filename = "p_hat300-1"
    n_mean = 10
    threshold = True


class Mutag0(GraphDataset):
    """First graph of the MUTAG dataset.

    The MUTAG dataset is from :cite:`debnath1991structure,kriege2012subgraph` and is available
    `here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__.

    **Graph:**

    .. |mutag_0| image:: ../../_static/graphs/MUTAG_0.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |mutag_0|

    Attributes:
        n_mean = 6
        threshold = False
        n_samples = 20000
        modes = 17
    """

    _data_filename = "MUTAG_0"
    n_mean = 6
    threshold = False


class Mutag1(GraphDataset):
    """Second graph of the MUTAG dataset.

    The MUTAG dataset is from :cite:`debnath1991structure,kriege2012subgraph` and is available
    `here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__.

    **Graph:**

    .. |mutag_1| image:: ../../_static/graphs/MUTAG_1.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |mutag_1|

    Attributes:
        n_mean = 6
        threshold = False
        n_samples = 20000
        modes = 13
    """

    _data_filename = "MUTAG_1"
    n_mean = 6
    threshold = False


class Mutag2(GraphDataset):
    """Third graph of the MUTAG dataset.

    The MUTAG dataset is from :cite:`debnath1991structure,kriege2012subgraph` and is available
    `here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__.

    **Graph:**

    .. |mutag_2| image:: ../../_static/graphs/MUTAG_2.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |mutag_2|

    Attributes:
        n_mean = 6
        threshold = False
        n_samples = 20000
        modes = 13
    """

    _data_filename = "MUTAG_2"
    n_mean = 6
    threshold = False


class Mutag3(GraphDataset):
    """Fourth graph of the MUTAG dataset.

    The MUTAG dataset is from :cite:`debnath1991structure,kriege2012subgraph` and is available
    `here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__.

    **Graph:**

    .. |mutag_3| image:: ../../_static/graphs/MUTAG_3.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |mutag_3|

    Attributes:
        n_mean = 6
        threshold = False
        n_samples = 20000
        modes = 19
    """

    _data_filename = "MUTAG_3"
    n_mean = 6
    threshold = False


# pylint: disable=abstract-method
class MoleculeDataset(Dataset, ABC):
    r"""Class for loading datasets of pre-generated samples from molecules.

    Attributes:
        w (array): normal mode frequencies of the electronic ground state (:math:`\mbox{cm}^{-1}`)
        wp (array): normal mode frequencies of the electronic excited state (:math:`\mbox{cm}^{-1}`)
        Ud (array): Duschinsky matrix
        delta (array): Displacement vector, with entries :math:`\delta_i=\sqrt{\omega'_i/\hbar}d_i`,
           and :math:`d_i` is the Duschinsky displacement
        T (float): temperature (Kelvin)
    """

    def __init__(self):
        super().__init__()
        self.w = scipy.sparse.load_npz(DATA_PATH + self._data_filename + "_w.npz").toarray()[0]
        self.wp = scipy.sparse.load_npz(DATA_PATH + self._data_filename + "_wp.npz").toarray()[0]
        self.Ud = scipy.sparse.load_npz(DATA_PATH + self._data_filename + "_Ud.npz").toarray()
        self.delta = scipy.sparse.load_npz(
            DATA_PATH + self._data_filename + "_delta.npz"
        ).toarray()[0]

    # pylint: disable=missing-docstring
    @property
    @abstractmethod
    def T(self) -> bool:
        pass


class Formic(MoleculeDataset):
    """Zero temperature formic acid.

    The molecular parameters are obtained from Ref. :cite:`huh2015boson`.

    **Molecule:**

    .. |formic| image:: ../../_static/formic.png
        :align: middle
        :width: 250px
        :target: javascript:void(0);

    |formic|

    Attributes:
        n_mean = 1.56
        threshold = False
        n_samples = 20000
        modes = 14
        T = 0
    """

    _data_filename = "formic"
    n_mean = 1.56
    threshold = False
    T = 0



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit class specification for the general Gaussian Boson Sampling class of circuits."""

from strawberryfields.program_utils import CircuitError, Command, group_operations
import strawberryfields.ops as ops

from .gaussian import GaussianSpecs


class GBSSpecs(GaussianSpecs):
    """Circuit specifications for the general GBS class of circuits."""

    short_name = "gbs"
    primitives = {
        # meta operations
        "All",
        "_New_modes",
        "_Delete",
        # state preparations
        "Vacuum",
        "Coherent",
        "Squeezed",
        "DisplacedSqueezed",
        "Thermal",
        "Gaussian",
        # measurements
        "MeasureHomodyne",
        "MeasureHeterodyne",
        "MeasureFock",
        "MeasureThreshold",
        # channels
        "LossChannel",
        "ThermalLossChannel",
        # single mode gates
        "Dgate",
        "Sgate",
        "Rgate",
        "Fouriergate",
        "BSgate",
    }
    decompositions = {"Xgate": {}, "Zgate": {}, "Fouriergate": {}, "S2gate": {}}

    def compile(self, seq, registers):
        """Try to arrange a quantum circuit into a form suitable for Gaussian boson sampling.

        This method checks whether the circuit can be implemented as a Gaussian boson sampling
        problem, i.e., if it is equivalent to a circuit A+B, where the sequence A only contains
        Gaussian operations, and B only contains Fock measurements.

        If the answer is yes, the circuit is arranged into the A+B order, and all the Fock
        measurements are combined into a single :class:`MeasureFock` operation.

        Args:
            seq (Sequence[Command]): quantum circuit to modify
            registers (Sequence[RegRefs]): quantum registers
        Returns:
            List[Command]: modified circuit
        Raises:
            CircuitError: the circuit does not correspond to GBS
        """
        A, B, C = group_operations(seq, lambda x: isinstance(x, ops.MeasureFock))

        # C should be empty
        if C:
            raise CircuitError("Operations following the Fock measurements.")

        # A should only contain Gaussian operations
        # (but this is already guaranteed by group_operations() and our primitive set)

        # without Fock measurements GBS is pointless
        if not B:
            raise CircuitError("GBS circuits must contain Fock measurements.")

        # there should be only Fock measurements in B
        measured = set()
        for cmd in B:
            if not isinstance(cmd.op, ops.MeasureFock):
                raise CircuitError("The Fock measurements are not consecutive.")

            # combine the Fock measurements
            temp = set(cmd.reg)
            if measured & temp:
                raise CircuitError("Measuring the same mode more than once.")
            measured |= temp

        # replace B with a single Fock measurement
        B = [Command(ops.MeasureFock(), sorted(list(measured), key=lambda x: x.ind))]
        return super().compile(A + B, registers)



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *
from strawberryfields.utils import scale
from numpy import pi, sqrt

# initialize engine and program objects
eng = sf.Engine(backend="gaussian")
teleportation = sf.Program(3)

with teleportation.context as q:
    psi, alice, bob = q[0], q[1], q[2]

    # state to be teleported:
    Coherent(1+0.5j) | psi

    # 50-50 beamsplitter
    BS = BSgate(pi/4, 0)

    # maximally entangled states
    Squeezed(-2) | alice
    Squeezed(2) | bob
    BS | (alice, bob)

    # Alice performs the joint measurement
    # in the maximally entangled basis
    BS | (psi, alice)
    MeasureX | psi
    MeasureP | alice

    # Bob conditionally displaces his mode
    # based on Alice's measurement result
    Xgate(psi.par*sqrt(2)) | bob
    Zgate(alice.par*sqrt(2)) | bob
    # end circuit

results = eng.run(teleportation)
# view Bob's output state and fidelity
print(results.samples)
print(results.state.displacement([2]))
print(results.state.fidelity_coherent([0, 0, 1+0.5j]))



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys

from setuptools import setup, find_packages


with open("strawberryfields/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")


requirements = [
    "numpy>=1.17.4",
    "scipy>=1.0.0",
    "sympy>=1.5",
    "networkx>=2.0",
    "quantum-blackbird>=0.2.3",
    "python-dateutil>=2.8.0",
    "thewalrus>=0.11",
    "numba",
    "toml",
    "appdirs",
    "requests>=2.22.0",
    "urllib3>=1.25.3",
]

info = {
    "name": "StrawberryFields",
    "version": version,
    "maintainer": "Xanadu Inc.",
    "maintainer_email": "software@xanadu.ai",
    "url": "https://github.com/XanaduAI/StrawberryFields",
    "license": "Apache License 2.0",
    "packages": find_packages(where="."),
    "package_data": {"strawberryfields": ["backends/data/*", "apps/data/*"]},
    "include_package_data": True,
    "entry_points" : {
        'console_scripts': [
            'sf=strawberryfields.cli:main'
        ]
    },
    "description": "Open source library for continuous-variable quantum computation",
    "long_description": open("README.rst", encoding="utf-8").read(),
    "provides": ["strawberryfields"],
    "install_requires": requirements,
    # 'extras_require': extra_requirements,
    "command_options": {
        "build_sphinx": {"version": ("setup.py", version), "release": ("setup.py", version)}
    },
}

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering :: Physics",
]

setup(classifiers=classifiers, **(info))



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module implements :class:`BaseEngine` and its subclasses that are responsible for
communicating quantum programs represented by :class:`.Program` objects
to a backend that could be e.g., a simulator or a hardware quantum processor.
One can think of each BaseEngine instance as a separate quantum computation.
"""
import abc
import collections.abc
import logging
import time
from typing import Optional

import numpy as np

from strawberryfields.api import Connection, Job, Result
from strawberryfields.program import Program

from .backends import load_backend
from .backends.base import BaseBackend, NotApplicableError

# for automodapi, do not include the classes that should appear under the top-level strawberryfields namespace
__all__ = ["BaseEngine", "LocalEngine"]

log = logging.getLogger(__name__)


class BaseEngine(abc.ABC):
    r"""Abstract base class for quantum program executor engines.

    Args:
        backend (str): backend short name
        backend_options (Dict[str, Any]): keyword arguments for the backend
    """

    def __init__(self, backend, backend_options=None):
        if backend_options is None:
            backend_options = {}

        #: str: short name of the backend
        self.backend_name = backend
        #: Dict[str, Any]: keyword arguments for the backend
        self.backend_options = backend_options.copy()  # dict is mutable
        #: List[Program]: list of Programs that have been run
        self.run_progs = []
        #: List[List[Number]]: latest measurement results, shape == (modes, shots)
        self.samples = None

        if isinstance(backend, str):
            self.backend_name = backend
            self.backend = load_backend(backend)
        elif isinstance(backend, BaseBackend):
            self.backend_name = backend.short_name
            self.backend = backend
        else:
            raise TypeError("backend must be a string or a BaseBackend instance.")

    @abc.abstractmethod
    def __str__(self):
        """String representation."""

    @abc.abstractmethod
    def reset(self, backend_options):
        r"""Re-initialize the quantum computation.

        Resets the state of the engine and the quantum circuit represented by the backend.

        * The original number of modes is restored.
        * All modes are reset to the vacuum state.
        * All registers of previously run Programs are cleared of measured values.
        * List of previously run Progams is cleared.

        Note that the reset does nothing to any Program objects in existence, beyond
        erasing the measured values.

        **Example:**

        .. code-block:: python

            # create a program
            prog = sf.Program(3)

            with prog.context as q:
                ops.Sgate(0.543) | q[1]
                ops.BSgate(0.6, 0.1) | (q[2], q[0])

            # create an engine
            eng = sf.Engine("gaussian")

        Running the engine with the above program will
        modify the state of the statevector simulator:

        >>> eng.run(prog)
        >>> eng.backend.is_vacuum()
        False

        Resetting the engine will return the backend simulator
        to the vacuum state:

        >>> eng.reset()
        >>> eng.backend.is_vacuum()
        True

        .. note:: The ``reset()`` method only applies to statevector backends.

        Args:
            backend_options (Dict[str, Any]): keyword arguments for the backend,
                updating (overriding) old values
        """
        self.backend_options.update(backend_options)
        for p in self.run_progs:
            p._clear_regrefs()
        self.run_progs.clear()
        self.samples = None

    def print_applied(self, print_fn=print):
        """Print all the Programs run since the backend was initialized.

        This will be blank until the first call to :meth:`~.LocalEngine.run`. The output may
        differ compared to :meth:`.Program.print`, due to backend-specific
        device compilation performed by :meth:`~.LocalEngine.run`.

        **Example:**

        .. code-block:: python

            # create a program
            prog = sf.Program(2)

            with prog.context as q:
                ops.S2gate(0.543) | (q[0], q[1])

            # create an engine
            eng = sf.Engine("gaussian")

        Initially, the engine will have applied no operations:

        >>> eng.print_applied()
        None

        After running the engine, we can now see the quantum
        operations that were applied:

        >>> eng.run(prog)
        >>> eng.print_applied()
        Run 0:
        BSgate(0.7854, 0) | (q[0], q[1])
        Sgate(0.543, 0) | (q[0])
        Sgate(0.543, 0).H | (q[1])
        BSgate(0.7854, 0).H | (q[0], q[1])

        Note that the :class:`~.S2gate` has been decomposed into
        single-mode squeezers and beamsplitters, which are supported
        by the ``'gaussian'`` backend.

        Subsequent program runs can also be viewed:

        .. code-block:: python

            # a second program
            prog2 = sf.Program(2)
            with prog2.context as q:
                ops.S2gate(0.543) | (q[0], q[1])

        >>> eng.run(prog2)
        >>> eng.print_applied()
        Run 0:
        BSgate(0.7854, 0) | (q[0], q[1])
        Sgate(0.543, 0) | (q[0])
        Sgate(0.543, 0).H | (q[1])
        BSgate(0.7854, 0).H | (q[0], q[1])
        Run 1:
        Dgate(0.06, 0) | (q[0])

        Args:
            print_fn (function): optional custom function to use for string printing.
        """
        for k, r in enumerate(self.run_progs):
            print_fn("Run {}:".format(k))
            r.print(print_fn)

    @abc.abstractmethod
    def _init_backend(self, init_num_subsystems):
        """Initialize the backend.

        Args:
            init_num_subsystems (int): number of subsystems the backend is initialized to
        """

    @abc.abstractmethod
    def _run_program(self, prog, **kwargs):
        """Execute a single program on the backend.

        This method should not be called directly.

        Args:
            prog (Program): program to run
        Returns:
            list[Command]: commands that were applied to the backend
        """

    def _run(self, program, *, args, compile_options, **kwargs):
        """Execute the given programs by sending them to the backend.

        If multiple Programs are given they will be executed sequentially as
        parts of a single computation.
        For each :class:`.Program` instance given as input, the following happens:

        * The Program instance is compiled for the target backend.
        * The compiled program is executed on the backend.
        * The measurement results of each subsystem (if any) are stored in the :class:`.RegRef`
          instances of the corresponding Program, as well as in :attr:`~BaseEngine.samples`.
        * The compiled program is appended to :attr:`~BaseEngine.run_progs`.

        Finally, the result of the computation is returned.

        Args:
            program (Program, Sequence[Program]): quantum programs to run
            args (Dict[str, Any]): values for the free parameters in the program(s) (if any)
            compile_options (Dict[str, Any]): keyword arguments for :meth:`.Program.compile`

        The ``kwargs`` keyword arguments are passed to the backend API calls via :meth:`Operation.apply`.

        Returns:
            Result: results of the computation
        """

        def _broadcast_nones(val, dim):
            """Helper function to ensure register values have same shape, even if not measured"""
            if val is None and dim > 1:
                return [None] * dim
            return val

        if not isinstance(program, collections.abc.Sequence):
            program = [program]

        kwargs.setdefault("shots", 1)
        # NOTE: by putting ``shots`` into keyword arguments, it allows for the
        # signatures of methods in Operations to remain cleaner, since only
        # Measurements need to know about shots

        prev = self.run_progs[-1] if self.run_progs else None  # previous program segment
        for p in program:
            if prev is None:
                # initialize the backend
                self._init_backend(p.init_num_subsystems)
            else:
                # there was a previous program segment
                if not p.can_follow(prev):
                    raise RuntimeError(
                        "Register mismatch: program {}, '{}'.".format(len(self.run_progs), p.name)
                    )

                # Copy the latest measured values in the RegRefs of p.
                # We cannot copy from prev directly because it could be used in more than one
                # engine.
                for k, v in enumerate(self.samples):
                    p.reg_refs[k].val = v

            # bind free parameters to their values
            p.bind_params(args)

            # if the program hasn't been compiled for this backend, do it now
            target = self.backend.circuit_spec
            if target is not None and p.target != target:
                p = p.compile(target, **compile_options)
            p.lock()

            self._run_program(p, **kwargs)
            shots = kwargs.get("shots", 1)
            self.samples = [_broadcast_nones(p.reg_refs[k].val, shots) for k in sorted(p.reg_refs)]
            self.run_progs.append(p)

            prev = p

        if self.samples is not None:
            return Result(np.array(self.samples).T)


class LocalEngine(BaseEngine):
    """Local quantum program executor engine.

    The Strawberry Fields engine is used to execute :class:`.Program` instances
    on the chosen local backend, and makes the results available via :class:`.Result`.

    **Example:**

    The following example creates a Strawberry Fields
    quantum :class:`~.Program` and runs it using an engine.

    .. code-block:: python

        # create a program
        prog = sf.Program(2)

        with prog.context as q:
            ops.S2gate(0.543) | (q[0], q[1])

    We initialize the engine with the name of the local backend,
    and can pass optional backend options.

    >>> eng = sf.Engine("fock", backend_options={"cutoff_dim": 5})

    The :meth:`~.LocalEngine.run` method is used to execute quantum
    programs on the attached backend, and returns a :class:`.Result`
    object containing the results of the execution.

    >>> results = eng.run(prog)

    Args:
        backend (str, BaseBackend): short name of the backend, or a pre-constructed backend instance
        backend_options (None, Dict[str, Any]): keyword arguments to be passed to the backend
    """

    def __init__(self, backend, *, backend_options=None):
        backend_options = backend_options or {}
        super().__init__(backend, backend_options)

    def __str__(self):
        return self.__class__.__name__ + "({})".format(self.backend_name)

    def reset(self, backend_options=None):
        backend_options = backend_options or {}
        super().reset(backend_options)
        self.backend.reset(**self.backend_options)
        # TODO should backend.reset and backend.begin_circuit be combined?

    def _init_backend(self, init_num_subsystems):
        self.backend.begin_circuit(init_num_subsystems, **self.backend_options)

    def _run_program(self, prog, **kwargs):
        applied = []
        for cmd in prog.circuit:
            try:
                # try to apply it to the backend
                # NOTE we could also handle storing measured vals here
                cmd.op.apply(cmd.reg, self.backend, **kwargs)
                applied.append(cmd)
            except NotApplicableError:
                # command is not applicable to the current backend type
                raise NotApplicableError(
                    "The operation {} cannot be used with {}.".format(cmd.op, self.backend)
                ) from None
            except NotImplementedError:
                # command not directly supported by backend API
                raise NotImplementedError(
                    "The operation {} has not been implemented in {} for the arguments {}.".format(
                        cmd.op, self.backend, kwargs
                    )
                ) from None
        return applied

    def run(self, program, *, args=None, compile_options=None, **kwargs):
        """Execute quantum programs by sending them to the backend.

        Args:
            program (Program, Sequence[Program]): quantum programs to run
            args (dict[str, Any]): values for the free parameters in the program(s) (if any)
            compile_options (None, Dict[str, Any]): keyword arguments for :meth:`.Program.compile`

        Keyword Args:
            shots (int): number of times the program measurement evaluation is repeated
            modes (None, Sequence[int]): Modes to be returned in the ``Result.state`` :class:`.BaseState` object.
                ``None`` returns all the modes (default). An empty sequence means no state object is returned.
            eval (bool): If False, the backend returns unevaluated tf.Tensors instead of
                numbers/arrays for returned measurement results and state (default: True). TF backend only.
            session (tf.Session): TensorFlow session, used when evaluating returned measurement results and state.
                TF backend only.
            feed_dict (dict[str, Any]): TensorFlow feed dictionary, used when evaluating returned measurement results
                and state. TF backend only.

        Returns:
            Result: results of the computation
        """
        args = args or {}
        compile_options = compile_options or {}
        temp_run_options = {}

        if isinstance(program, collections.abc.Sequence):
            # succesively update all run option defaults.
            # the run options of successive programs
            # overwrite the run options of previous programs
            # in the list
            program_lst = program
            for p in program:
                temp_run_options.update(p.run_options)
        else:
            # single program to execute
            program_lst = [program]
            temp_run_options.update(program.run_options)

        temp_run_options.update(kwargs or {})
        temp_run_options.setdefault("shots", 1)
        temp_run_options.setdefault("modes", None)

        # avoid unexpected keys being sent to Operations
        eng_run_keys = ["eval", "session", "feed_dict", "shots"]
        eng_run_options = {
            key: temp_run_options[key] for key in temp_run_options.keys() & eng_run_keys
        }

        # check that batching is not used together with shots > 1
        if self.backend_options.get("batch_size", 0) and eng_run_options["shots"] > 1:
            raise NotImplementedError("Batching cannot be used together with multiple shots.")

        # check that post-selection and feed-forwarding is not used together with shots > 1
        for p in program_lst:
            for c in p.circuit:
                try:
                    if c.op.select and eng_run_options["shots"] > 1:
                        raise NotImplementedError(
                            "Post-selection cannot be used together with multiple shots."
                        )
                except AttributeError:
                    pass

                if c.op.measurement_deps and eng_run_options["shots"] > 1:
                    raise NotImplementedError(
                        "Feed-forwarding of measurements cannot be used together with multiple shots."
                    )

        result = super()._run(
            program, args=args, compile_options=compile_options, **eng_run_options
        )

        modes = temp_run_options["modes"]

        if modes is None or modes:
            # state object requested
            # session and feed_dict are needed by TF backend both during simulation (if program
            # contains measurements) and state object construction.
            result._state = self.backend.state(**temp_run_options)

        return result


class RemoteEngine:
    """A quantum program executor engine that provides a simple interface for
    running remote jobs in a blocking or non-blocking manner.

    **Example:**

    The following examples instantiate an engine with the default configuration, and
    run both blocking and non-blocking jobs.

    Run a blocking job:

    >>> engine = RemoteEngine("X8_01")
    >>> result = engine.run(program, shots=1) # blocking call
    >>> result
    [[0 1 0 2 1 0 0 0]]

    Run a non-blocking job:

    >>> job = engine.run_async(program, shots=1)
    >>> job.status
    "queued"
    >>> job.result
    InvalidJobOperationError
    >>> job.refresh()
    >>> job.status
    "complete"
    >>> job.result
    [[0 1 0 2 1 0 0 0]]

    Args:
        target (str): the target device
        connection (strawberryfields.api.Connection): a connection to the remote job
            execution platform
        backend_options (Dict[str, Any]): keyword arguments for the backend
    """

    POLLING_INTERVAL_SECONDS = 1
    VALID_TARGETS = ("X8_01", "X12_01", "X12_02")
    DEFAULT_TARGETS = {"X8": "X8_01", "X12": "X12_01"}

    def __init__(
        self, target: str, connection: Connection = Connection(), backend_options: dict = None
    ):
        self._target = self.DEFAULT_TARGETS.get(target, target)

        if self._target not in self.VALID_TARGETS:
            raise ValueError(
                "Invalid engine target: {} (valid targets: {})".format(
                    target, tuple(self.DEFAULT_TARGETS.keys()) + self.VALID_TARGETS
                )
            )

        self._connection = connection
        self._backend_options = backend_options or {}

    @property
    def target(self) -> str:
        """The target device used by the engine.

        Returns:
            str: the name of the target
        """
        return self._target

    @property
    def connection(self) -> Connection:
        """The connection object used by the engine.

        Returns:
            strawberryfields.api.Connection
        """
        return self._connection

    def run(self, program: Program, *, compile_options=None, **kwargs) -> Optional[Result]:
        """Runs a blocking job.

        In the blocking mode, the engine blocks until the job is completed, failed, or
        cancelled. A job in progress can be cancelled with a keyboard interrupt (`ctrl+c`).

        If the job completes successfully, the result is returned; if the job
        fails or is cancelled, ``None`` is returned.

        Args:
            program (strawberryfields.Program): the quantum circuit

        Keyword Args:
            shots (Optional[int]): The number of shots for which to run the job. If this
                argument is not provided, the shots are derived from the given ``program``.

        Returns:
            [strawberryfields.api.Result, None]: the job result if successful, and
            ``None`` otherwise
        """
        job = self.run_async(program, compile_options=compile_options, **kwargs)
        try:
            while True:
                job.refresh()
                if job.status == "complete":
                    return job.result
                if job.status == "failed":
                    log.warning(
                        "The remote job failed due to an internal server error; "
                        "please try again."
                    )
                    return None
                time.sleep(self.POLLING_INTERVAL_SECONDS)
        except KeyboardInterrupt:
            self._connection.cancel_job(job.id)
            return None

    def run_async(self, program: Program, *, compile_options=None, **kwargs) -> Job:
        """Runs a non-blocking remote job.

        In the non-blocking mode, a ``Job`` object is returned immediately, and the user can
        manually refresh the status and check for updated results of the job.

        Args:
            program (strawberryfields.Program): the quantum circuit
            compile_options (None, Dict[str, Any]): keyword arguments for :meth:`.Program.compile`

        Keyword Args:
            shots (Optional[int]): The number of shots for which to run the job. If this
                argument is not provided, the shots are derived from the given ``program``.

        Returns:
            strawberryfields.api.Job: the created remote job
        """
        # get the specific chip to submit the program to
        # TODO: this should be provided by the chip API, rather
        # than built-in to Strawberry Fields.
        compile_options = compile_options or {}
        kwargs.update(self._backend_options)

        if program.target is None or (program.target.split("_")[0] != self.target.split("_")[0]):
            # Program is either:
            #
            # * uncompiled (program.target is None)
            # * compiled to a different chip family to the engine target
            #
            # In both cases, recompile the program to match the intended target.
            program = program.compile(self.target, **compile_options)

        # update the run options if provided
        run_options = {}
        run_options.update(program.run_options)
        run_options.update(kwargs or {})

        if "shots" not in run_options:
            raise ValueError("Number of shots must be specified.")

        return self._connection.create_job(self.target, program, run_options)

    def __repr__(self):
        return "<{}: target={}, connection={}>".format(
            self.__class__.__name__, self.target, self.connection
        )

    def __str__(self):
        return self.__repr__()


class Engine(LocalEngine):
    """dummy"""

    # alias for backwards compatibility
    __doc__ = LocalEngine.__doc__



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module contains functions used to load, store, save, and modify
configuration options for Strawberry Fields.
"""
import logging as log
import os

import toml
from appdirs import user_config_dir

log.getLogger()

DEFAULT_CONFIG_SPEC = {
    "api": {
        "authentication_token": (str, ""),
        "hostname": (str, "platform.strawberryfields.ai"),
        "use_ssl": (bool, True),
        "port": (int, 443),
    }
}


class ConfigurationError(Exception):
    """Exception used for configuration errors"""


def load_config(filename="config.toml", **kwargs):
    """Load configuration from keyword arguments, configuration file or
    environment variables.

    .. note::

        The configuration dictionary will be created based on the following
        (order defines the importance, going from most important to least
        important):

        1. keyword arguments passed to ``load_config``
        2. data contained in environmental variables (if any)
        3. data contained in a configuration file (if exists)

    Keyword Args:
        filename (str): the name of the configuration file to look for.
            Additional configuration options are detailed in
            :doc:`/code/sf_configuration`

    Returns:
        dict[str, dict[str, Union[str, bool, int]]]: the configuration
    """
    config = create_config()

    filepath = get_config_filepath(filename=filename)

    if filepath is not None:
        loaded_config = load_config_file(filepath)
        valid_api_options = keep_valid_options(loaded_config["api"])
        config["api"].update(valid_api_options)
    else:
        log.info("No Strawberry Fields configuration file found.")

    update_from_environment_variables(config)

    valid_kwargs_config = keep_valid_options(kwargs)
    config["api"].update(valid_kwargs_config)

    return config


def create_config(authentication_token=None, **kwargs):
    """Create a configuration object that stores configuration related data
    organized into sections.

    The configuration object contains API-related configuration options. This
    function takes into consideration only pre-defined options.

    If called without passing any keyword arguments, then a default
    configuration object is created.

    Keyword Args:
        Configuration options as detailed in :doc:`/code/sf_configuration`

    Returns:
        dict[str, dict[str, Union[str, bool, int]]]: the configuration
            object
    """
    authentication_token = authentication_token or ""
    hostname = kwargs.get("hostname", DEFAULT_CONFIG_SPEC["api"]["hostname"][1])
    use_ssl = kwargs.get("use_ssl", DEFAULT_CONFIG_SPEC["api"]["use_ssl"][1])
    port = kwargs.get("port", DEFAULT_CONFIG_SPEC["api"]["port"][1])

    config = {
        "api": {
            "authentication_token": authentication_token,
            "hostname": hostname,
            "use_ssl": use_ssl,
            "port": port,
        }
    }
    return config


def get_config_filepath(filename="config.toml"):
    """Get the filepath of the first configuration file found from the defined
    configuration directories (if any).

    .. note::

        The following directories are checked (in the following order):

        * The current working directory
        * The directory specified by the environment variable ``SF_CONF`` (if specified)
        * The user configuration directory (if specified)

    Keyword Args:
        filename (str): the configuration file to look for

    Returns:
         Union[str, None]: the filepath to the configuration file or None, if
             no file was found
    """
    current_dir = os.getcwd()
    sf_env_config_dir = os.environ.get("SF_CONF", "")
    sf_user_config_dir = user_config_dir("strawberryfields", "Xanadu")

    directories = [current_dir, sf_env_config_dir, sf_user_config_dir]
    for directory in directories:
        filepath = os.path.join(directory, filename)
        if os.path.exists(filepath):
            return filepath

    return None


def load_config_file(filepath):
    """Load a configuration object from a TOML formatted file.

    Args:
        filepath (str): path to the configuration file

    Returns:
         dict[str, dict[str, Union[str, bool, int]]]: the configuration
            object that was loaded
    """
    with open(filepath, "r") as f:
        config_from_file = toml.load(f)
    return config_from_file


def keep_valid_options(sectionconfig):
    """Filters the valid options in a section of a configuration dictionary.

    Args:
        sectionconfig (dict[str, Union[str, bool, int]]): the section of the
            configuration to check

    Returns:
        dict[str, Union[str, bool, int]]: the keep section of the
            configuration
    """
    return {k: v for k, v in sectionconfig.items() if k in VALID_KEYS}


def update_from_environment_variables(config):
    """Updates the current configuration object from data stored in environment
    variables.

    The list of environment variables can be found at :mod:`strawberryfields.configuration`

    Args:
        config (dict[str, dict[str, Union[str, bool, int]]]): the
            configuration to be updated
    Returns:
        dict[str, dict[str, Union[str, bool, int]]]): the updated
        configuration
    """
    for section, sectionconfig in config.items():
        env_prefix = "SF_{}_".format(section.upper())
        for key in sectionconfig:
            env = env_prefix + key.upper()
            if env in os.environ:
                config[section][key] = parse_environment_variable(key, os.environ[env])


def parse_environment_variable(key, value):
    """Parse a value stored in an environment variable.

    Args:
        key (str): the name of the environment variable
        value (Union[str, bool, int]): the value obtained from the environment
            variable

    Returns:
        [str, bool, int]: the parsed value
    """
    trues = (True, "true", "True", "TRUE", "1", 1)
    falses = (False, "false", "False", "FALSE", "0", 0)

    if DEFAULT_CONFIG_SPEC["api"][key][0] is bool:
        if value in trues:
            return True

        if value in falses:
            return False

        raise ValueError("Boolean could not be parsed")

    if DEFAULT_CONFIG_SPEC["api"][key][0] is int:
        return int(value)

    return value


def store_account(authentication_token, filename="config.toml", location="user_config", **kwargs):
    r"""Configure Strawberry Fields for access to the Xanadu cloud platform by
    saving your account credentials.

    The configuration file can be created in the following locations:

    - A global user configuration directory (``"user_config"``)
    - The current working directory (``"local"``)

    This global user configuration directory differs depending on the operating system:

    * On Linux: ``~/.config/strawberryfields``
    * On Windows: ``C:\Users\USERNAME\AppData\Local\Xanadu\strawberryfields``
    * On MacOS: ``~/Library/Application Support/strawberryfields``

    By default, Strawberry Fields will load the configuration and account credentials from the global
    user configuration directory, no matter the working directory. However, if there exists a configuration
    file in the *local* working directory, this takes precedence. The ``"local"`` option is therefore useful
    for maintaining per-project configuration settings.

    **Examples:**

    In these examples ``"AUTHENTICATION_TOKEN"`` should be replaced with a valid authentication
    token.

    Access to the Xanadu cloud can be configured as follows:

    >>> sf.store_account("AUTHENTICATION_TOKEN")

    This creates the following ``"config.toml"`` file:

    .. code-block:: toml

        [api]
        authentication_token = "AUTHENTICATION_TOKEN"
        hostname = "platform.strawberryfields.ai"
        use_ssl = true
        port = 443

    You can also create the configuration file locally (in the **current
    working directory**) the following way:

    >>> import strawberryfields as sf
    >>> sf.store_account("AUTHENTICATION_TOKEN", location="local")

    Each of the configuration options can be passed as further keyword
    arguments as well (see the :doc:`/code/sf_configuration` page
    for a list of options):

    >>> import strawberryfields as sf
    >>> sf.store_account("AUTHENTICATION_TOKEN", location="local", hostname="MyHost", use_ssl=False, port=123)

    This creates the following ``"config.toml"`` file in the **current working directory**:

    .. code-block:: toml

        [api]
        authentication_token = "AUTHENTICATION_TOKEN"
        hostname = "MyHost"
        use_ssl = false
        port = 123

    Args:
        authentication_token (str): API token for authentication to the Xanadu cloud platform.
            This is required for submitting remote jobs using :class:`~.RemoteEngine`.

    Keyword Args:
        location (str): determines where the configuration file should be saved
        filename (str): the name of the configuration file to look for

    Additional configuration options are detailed in :doc:`/code/sf_configuration` and can be passed
    as keyword arguments.
    """
    if location == "user_config":
        directory = user_config_dir("strawberryfields", "Xanadu")

        # Create target Directory if it doesn't exist
        os.makedirs(directory, exist_ok=True)
    elif location == "local":
        directory = os.getcwd()
    else:
        raise ConfigurationError("This location is not recognized.")

    filepath = os.path.join(directory, filename)

    config = create_config(authentication_token=authentication_token, **kwargs)
    save_config_to_file(config, filepath)


def save_config_to_file(config, filepath):
    """Saves a configuration to a TOML file.

    Args:
        config (dict[str, dict[str, Union[str, bool, int]]]): the
            configuration to be saved
        filepath (str): path to the configuration file
    """
    with open(filepath, "w") as f:
        toml.dump(config, f)


VALID_KEYS = set(create_config()["api"].keys())
DEFAULT_CONFIG = create_config()
configuration = load_config()
config_filepath = get_config_filepath()



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fock backend simulator"""
# pylint: disable=protected-access,too-many-public-methods

import string
from cmath import phase
import numpy as np

from strawberryfields.backends import BaseFock, ModeMap
from strawberryfields.backends.states import BaseFockState

from .circuit import Circuit

indices = string.ascii_lowercase


class FockBackend(BaseFock):
    r"""Implements a simulation of quantum optical circuits in a truncated
    Fock basis using NumPy, returning a :class:`~.BaseFock`
    state object.

    The primary component of the FockBackend is a
    :attr:`~.FockBackend.circuit` object which is used to simulate a multi-mode quantum optical system. The
    :class:`FockBackend` provides the basic API-compatible interface to the simulator, while the
    :attr:`~.FockBackend.circuit` object actually carries out the mathematical simulation.

    The :attr:`~.FockBackend.circuit` simulator maintains an internal tensor representation of the quantum state of a multi-mode quantum optical system
    using a (truncated) Fock basis representation. As its various state manipulation methods are called, the quantum state is updated
    to reflect these changes. The simulator will try to keep the internal state in a pure (vector) representation
    for as long as possible. Unitary gates will not change the type of representation, while state preparations and measurements will.

    A number of factors determine the shape and dimensionality of the state tensor:

    * the underlying state representation being used (either a ket vector or a density matrix)
    * the number of modes :math:`n` actively being simulated
    * the cutoff dimension :math:`D` for the Fock basis

    The state tensor corresponds to a multimode quantum system. If the
    representation is a pure state, the state tensor has shape
    :math:`(\underbrace{D,...,D}_{n~\text{times}})`.
    In a mixed state representation, the state tensor has shape
    :math:`(\underbrace{D,D,...,D,D}_{2n~\text{times}})`.
    Indices for the same mode appear consecutively. Hence, for a mixed state, the first two indices
    are for the first mode, the second are for the second mode, etc.

    .. 
        .. currentmodule:: strawberryfields.backends.fockbackend
        .. autosummary::
            :toctree:

            ~circuit.Circuit
            ~ops
    """

    short_name = "fock"
    circuit_spec = "fock"

    def __init__(self):
        """Instantiate a FockBackend object."""
        super().__init__()
        self._supported["mixed_states"] = True
        self._init_modes = None  #: int: initial number of modes in the circuit
        self._modemap = None  #: Modemap: maps external mode indices to internal ones
        self.circuit = (
            None  #: ~.fockbackend.circuit.Circuit: representation of the simulated quantum state
        )

    def _remap_modes(self, modes):
        if isinstance(modes, int):
            modes = [modes]
            was_int = True
        else:
            was_int = False
        map_ = self._modemap.show()
        submap = [map_[m] for m in modes]
        if not self._modemap.valid(modes) or None in submap:
            raise ValueError("The specified modes are not valid.")

        remapped_modes = self._modemap.remap(modes)
        if was_int:
            remapped_modes = remapped_modes[0]
        return remapped_modes

    def begin_circuit(self, num_subsystems, **kwargs):
        r"""Instantiate a quantum circuit.

        Instantiates a representation of a quantum optical state with ``num_subsystems`` modes.
        The state is initialized to vacuum.

        The modes in the circuit are indexed sequentially using integers, starting from zero.
        Once an index is assigned to a mode, it can never be re-assigned to another mode.
        If the mode is deleted its index becomes invalid.
        An operation acting on an invalid or unassigned mode index raises an ``IndexError`` exception.

        Args:
            num_subsystems (int): number of modes in the circuit

        Keyword Args:
            cutoff_dim (int): Numerical Hilbert space cutoff dimension for the modes.
                For each mode, the simulator can represent the Fock states :math:`\ket{0}, \ket{1}, \ldots, \ket{\text{cutoff_dim}-1}`.
            pure (bool): If True (default), use a pure state representation (otherwise will use a mixed state representation).
        """
        cutoff_dim = kwargs.get("cutoff_dim", None)
        pure = kwargs.get("pure", True)
        if cutoff_dim is None:
            raise ValueError("Argument 'cutoff_dim' must be passed to the Fock backend")
        if not isinstance(cutoff_dim, int):
            raise ValueError("Argument 'cutoff_dim' must be a positive integer")
        if not isinstance(num_subsystems, int):
            raise ValueError("Argument 'num_subsystems' must be a positive integer")
        if not isinstance(pure, bool):
            raise ValueError("Argument 'pure' must be either True or False")

        self._init_modes = num_subsystems
        self.circuit = Circuit(num_subsystems, cutoff_dim, pure)
        self._modemap = ModeMap(num_subsystems)

    def add_mode(self, n=1):
        self.circuit.alloc(n)
        self._modemap.add(n)

    def del_mode(self, modes):
        remapped_modes = self._remap_modes(modes)
        if isinstance(remapped_modes, int):
            remapped_modes = [remapped_modes]
        self.circuit.dealloc(remapped_modes)
        self._modemap.delete(modes)

    def get_modes(self):
        return [i for i, j in enumerate(self._modemap._map) if j is not None]

    def reset(self, pure=True, **kwargs):
        cutoff = kwargs.get("cutoff_dim", self.circuit._trunc)
        self._modemap.reset()
        self.circuit.reset(pure, num_subsystems=self._init_modes, cutoff_dim=cutoff)

    def prepare_vacuum_state(self, mode):
        self.circuit.prepare_mode_fock(0, self._remap_modes(mode))

    def prepare_coherent_state(self, alpha, mode):
        self.circuit.prepare_mode_coherent(alpha, self._remap_modes(mode))

    def prepare_squeezed_state(self, r, phi, mode):
        self.circuit.prepare_mode_squeezed(r, phi, self._remap_modes(mode))

    def prepare_displaced_squeezed_state(self, alpha, r, phi, mode):
        self.circuit.prepare_mode_displaced_squeezed(alpha, r, phi, self._remap_modes(mode))

    def prepare_thermal_state(self, nbar, mode):
        self.circuit.prepare_mode_thermal(nbar, self._remap_modes(mode))

    def rotation(self, phi, mode):
        self.circuit.phase_shift(phi, self._remap_modes(mode))

    def displacement(self, alpha, mode):
        self.circuit.displacement(alpha, self._remap_modes(mode))

    def squeeze(self, z, mode):
        self.circuit.squeeze(abs(z), phase(z), self._remap_modes(mode))

    def two_mode_squeeze(self, z, mode1, mode2):
        self.circuit.two_mode_squeeze(
            abs(z), phase(z), self._remap_modes(mode1), self._remap_modes(mode2)
        )

    def beamsplitter(self, t, r, mode1, mode2):
        if isinstance(t, complex):
            raise ValueError("Beamsplitter transmittivity t must be a float.")
        self.circuit.beamsplitter(
            t, abs(r), phase(r), self._remap_modes(mode1), self._remap_modes(mode2)
        )

    def measure_homodyne(self, phi, mode, shots=1, select=None, **kwargs):
        """Perform a homodyne measurement on the specified mode.

        See :meth:`.BaseBackend.measure_homodyne`.

        Keyword Args:
            num_bins (int): Number of equally spaced bins for the probability distribution function
                (pdf) simulating the homodyne measurement (default: 100000).
            max (float): The pdf is discretized onto the 1D grid [-max,max] (default: 10).
        """
        if shots != 1:
            raise NotImplementedError(
                "fock backend currently does not support " "shots != 1 for homodyne measurement"
            )
        return self.circuit.measure_homodyne(phi, self._remap_modes(mode), select=select, **kwargs)

    def loss(self, T, mode):
        self.circuit.loss(T, self._remap_modes(mode))

    def is_vacuum(self, tol=0.0, **kwargs):
        return self.circuit.is_vacuum(tol)

    def get_cutoff_dim(self):
        return self.circuit._trunc

    def state(self, modes=None, **kwargs):
        s, pure = self.circuit.get_state()

        if modes is None:
            # reduced state is full state
            red_state = s
            num_modes = len(s.shape) if pure else len(s.shape) // 2
            modes = [m for m in range(num_modes)]
        else:
            # convert to mixed state representation
            if pure:
                num_modes = len(s.shape)
                left_str = [indices[i] for i in range(0, 2 * num_modes, 2)]
                right_str = [indices[i] for i in range(1, 2 * num_modes, 2)]
                out_str = [indices[: 2 * num_modes]]
                einstr = "".join(left_str + [","] + right_str + ["->"] + out_str)
                rho = np.einsum(einstr, s, s.conj())
            else:
                rho = s

            # reduce rho down to specified subsystems
            if isinstance(modes, int):
                modes = [modes]

            if len(modes) != len(set(modes)):
                raise ValueError("The specified modes cannot be duplicated.")

            num_modes = len(rho.shape) // 2
            if len(modes) > num_modes:
                raise ValueError(
                    "The number of specified modes cannot be larger than the number of subsystems."
                )

            keep_indices = indices[: 2 * len(modes)]
            trace_indices = indices[2 * len(modes) : len(modes) + num_modes]
            ind = [i * 2 for i in trace_indices]
            ctr = 0

            for m in range(num_modes):
                if m in modes:
                    ind.insert(m, keep_indices[2 * ctr : 2 * (ctr + 1)])
                    ctr += 1

            indStr = "".join(ind) + "->" + keep_indices
            red_state = np.einsum(indStr, rho)

            # permute indices of returned state to reflect the ordering of modes (we know and hence can assume that red_state is a mixed state)
        if modes != sorted(modes):
            mode_permutation = np.argsort(modes)
            index_permutation = [2 * x + i for x in mode_permutation for i in (0, 1)]
            red_state = np.transpose(red_state, np.argsort(index_permutation))

        cutoff = self.circuit._trunc
        mode_names = ["q[{}]".format(i) for i in np.array(self.get_modes())[modes]]
        state = BaseFockState(red_state, len(modes), pure, cutoff, mode_names)
        return state

    # ==============================================
    # Fock state specific
    # ==============================================

    def prepare_fock_state(self, n, mode):
        self.circuit.prepare_mode_fock(n, self._remap_modes(mode))

    def prepare_ket_state(self, state, modes):
        self.circuit.prepare_multimode(state, self._remap_modes(modes))

    def prepare_dm_state(self, state, modes):
        self.circuit.prepare_multimode(state, self._remap_modes(modes))

    def cubic_phase(self, gamma, mode):
        self.circuit.cubic_phase_shift(gamma, self._remap_modes(mode))

    def kerr_interaction(self, kappa, mode):
        self.circuit.kerr_interaction(kappa, self._remap_modes(mode))

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        self.circuit.cross_kerr_interaction(
            kappa, self._remap_modes(mode1), self._remap_modes(mode2)
        )

    def measure_fock(self, modes, shots=1, select=None, **kwargs):
        if shots != 1:
            raise NotImplementedError(
                "fock backend currently does not support " "shots != 1 for Fock measurement"
            )
        return self.circuit.measure_fock(self._remap_modes(modes), select=select)



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The classes in this module represent parameters passed to the
quantum operations represented by :class:`~.Operation` subclasses.

Parameter types
---------------

There are three basic types of parameters:

1. **Numerical parameters** (bound and fixed): An immediate, immutable numerical object
   (float, complex, int, numerical array).
   Implemented as-is, not encapsulated in a class.

2. **Measured parameters** (bound but not fixed): Certain quantum circuits/protocols require that
   Operations can be conditioned on measurement results obtained during the execution of the
   circuit. In this case the parameter value is not known/fixed until the measurement is made
   (or simulated). Represented by :class:`MeasuredParameter` instances.
   Constructed from the :class:`.RegRef` instance storing the measurement
   result using the :meth:`.RegRef.par` method.

3. **Free parameters** (not bound nor fixed): A *parametrized circuit template* is a circuit that
   depends on a number of unbound (free) parameters. These parameters need to be bound to fixed
   numerical values before the circuit can be executed on a hardware quantum device or a numeric
   simulator. Represented by :class:`FreeParameter` instances.
   Simulators with symbolic capability can accept a parametrized circuit as input (and should
   return symbolic expressions representing the measurement results, with the same free parameters,
   as output).
   Free parameters belong to a single :class:`.Program` instance, are constructed using the
   :meth:`.Program.params` method, and are bound using :meth:`.Program.bind_params`.

:class:`.Operation` subclass constructors accept parameters that are functions or algebraic
combinations of any number of these basic parameter types. This is made possible by
:class:`MeasuredParameter` and :class:`FreeParameter` inheriting from :class:`sympy.Symbol`.

.. note:: Binary arithmetic operations between sympy symbols and numpy arrays produces numpy object arrays containing sympy symbols.


Operation lifecycle
-------------------

The normal lifecycle of an Operation object and its associated parameters is as follows:

* An Operation instance is constructed, and given some input arguments.
  In :meth:`.Operation.__init__`,
  the RegRef dependencies of measured parameters are added to :attr:`.Operation._measurement_deps`.

* The Operation instance is applied using its :meth:`~ops.Operation.__or__`
  method inside a :class:`.Program` context.
  This creates a :class:`.Command` instance that wraps
  the Operation and the RegRefs it acts on, which is appended to :attr:`.Program.circuit`.

* Before the Program is run, it is compiled and optimized for a specific backend. This involves
  checking that the Program only contains valid Operations, decomposing non-elementary Operations
  using :meth:`~ops.Operation.decompose`, and finally merging and commuting Commands inside
  the graph representing the quantum circuit.
  The circuit graph is built using the knowledge of which subsystems the Commands act and depend on.

* Decompositions, merges, and commutations often involve the creation of new Operations with algebraically
  transformed parameters.
  For example, merging two :class:`.Gate` instances of the same subclass involves
  adding their first parameters after equality-comparing the others. This is easily done if
  all the parameters have an immediate numerical value.
  Measured and free parameters are handled symbolically by Sympy.

* The compiled Program is run by a :class:`.BaseEngine` instance, which calls the
  :meth:`~ops.Operation.apply` method of each Operation in turn.

* :meth:`~ops.Operation.apply` then calls :meth:`~ops.Operation._apply` which is redefined by each Operation subclass.
  It evaluates the value of the parameters using :func:`par_evaluate`, and
  may perform additional numeric transformations on them.
  The parameter values are finally passed to the appropriate backend API method.
  It is up to the backend to either accept NumPy arrays and Tensorflow objects as parameters, or not.


What we cannot do at the moment:

* Use anything except integers and RegRefs (or Sequences thereof) as the subsystem argument
  for the :meth:`~ops.Operation.__or__` method.
  Technically we could allow any parameters that evaluate into an integer.
"""
# pylint: disable=too-many-ancestors,unused-argument,protected-access

import collections.abc
import functools
import types

import numpy as np
import sympy
import sympy.functions as sf


def wrap_mathfunc(func):
    """Applies the wrapped sympy function elementwise to NumPy arrays.

    Required because the sympy math functions cannot deal with NumPy arrays.
    We implement no broadcasting; if the first argument is a NumPy array, we assume
    all the arguments are arrays of the same shape.
    """

    @functools.wraps(func)
    def wrapper(*args):
        temp = [isinstance(k, np.ndarray) for k in args]
        if any(temp):
            if not all(temp):
                raise ValueError(
                    "Parameter functions with array arguments: all the arguments must be arrays of the same shape."
                )
            for k in args[1:]:
                if len(k) != len(args[0]):
                    raise ValueError(
                        "Parameter functions with array arguments: all the arguments must be arrays of the same shape."
                    )
            # apply func elementwise, recursively, on the args
            return np.array([wrapper(*k) for k in zip(*args)])
        return func(*args)

    return wrapper


par_funcs = types.SimpleNamespace(
    **{name: wrap_mathfunc(getattr(sf, name)) for name in dir(sf) if name[0] != "_"}
)
"""SimpleNamespace: Namespace of mathematical functions for manipulating Parameters.
Consists of all :mod:`sympy.functions` public members, which we wrap with :func:`wrap_mathfunc`.
"""


class ParameterError(RuntimeError):
    """Exception raised when the Parameter classes encounter an illegal operation.

    E.g., trying to use a measurement result before it is available.
    """


def is_object_array(p):
    """Returns True iff p is an object array.

    Args:
        p (Any): object to be checked

    Returns:
        bool: True iff p is a NumPy object array
    """
    return isinstance(p, np.ndarray) and p.dtype == object


def par_evaluate(params):
    """Evaluate an Operation parameter sequence.

    Any parameters descending from :class:`sympy.Basic` are evaluated, others are returned as-is.
    Evaluation means that free and measured parameters are replaced by their numeric values.
    NumPy object arrays are evaluated elementwise.

    Alternatively, evaluates a single parameter and returns its value.

    Args:
        params (Sequence[Any]): parameters to evaluate

    Returns:
        list[Any]: evaluated parameters
    """
    scalar = False
    if not isinstance(params, collections.abc.Sequence):
        scalar = True
        params = [params]

    def do_evaluate(p):
        """Evaluates a single parameter."""
        if is_object_array(p):
            return np.array([do_evaluate(k) for k in p])

        if not par_is_symbolic(p):
            return p

        # using lambdify we can also substitute np.ndarrays and tf.Tensors for the atoms
        atoms = list(p.atoms(MeasuredParameter, FreeParameter))
        # evaluate the atoms of the expression
        vals = [k._eval_evalf(None) for k in atoms]
        # use the tensorflow printer if any of the symbolic parameter values are TF objects
        # (we do it like this to avoid importing tensorflow if it's not needed)
        is_tf = (type(v).__module__.startswith("tensorflow") for v in vals)
        printer = "tensorflow" if any(is_tf) else "numpy"
        func = sympy.lambdify(atoms, p, printer)
        return func(*vals)

    ret = list(map(do_evaluate, params))
    if scalar:
        return ret[0]
    return ret


def par_is_symbolic(p):
    """Returns True iff p is a symbolic Operation parameter instance.

    If a parameter inherits :class:`sympy.Basic` it is symbolic.
    A NumPy object array is symbolic if any of its elements are.
    All other objects are considered not symbolic parameters.

    Note that :data:`strawberryfields.math` functions applied to numerical (non-symbolic) parameters return
    symbolic parameters.
    """
    if is_object_array(p):
        return any(par_is_symbolic(k) for k in p)
    return isinstance(p, sympy.Basic)


def par_convert(args, prog):
    """Convert Blackbird symbolic Operation arguments into their SF counterparts.

    Args:
        args (Iterable[Any]): Operation arguments
        prog (Program): program containing the Operations.

    Returns:
        list[Any]: converted arguments
    """

    def do_convert(a):
        if isinstance(a, sympy.Basic):
            # substitute SF symbolic parameter objects for Blackbird ones
            s = {}
            for k in a.atoms(sympy.Symbol):
                if k.name[0] == "q":
                    s[k] = MeasuredParameter(prog.register[int(k.name[1:])])
                else:
                    s[k] = prog.params(k.name)  # free parameter
            return a.subs(s)
        return a  # return non-symbols as-is

    return [do_convert(a) for a in args]


def par_regref_deps(p):
    """RegRef dependencies of an Operation parameter.

    Returns the RegRefs that the parameter depends on through the :class:`MeasuredParameter`
    atoms it contains.

    Args:
        p (Any): Operation parameter

    Returns:
        set[RegRef]: RegRefs the parameter depends on
    """
    ret = set()
    if is_object_array(p):
        # p is an object array, possibly containing symbols
        for k in p:
            ret.update(par_regref_deps(k))
    elif isinstance(p, sympy.Basic):
        # p is a Sympy expression, possibly containing measured parameters
        for k in p.atoms(MeasuredParameter):
            ret.add(k.regref)
    return ret


def par_str(p):
    """String representation of the Operation parameter.

    Args:
        p (Any): Operation parameter

    Returns:
        str: string representation
    """
    if isinstance(p, np.ndarray):
        np.set_printoptions(precision=4)
        return str(p)
    if par_is_symbolic(p):
        return str(p)
    return "{:.4g}".format(p)  # scalar parameters


class MeasuredParameter(sympy.Symbol):
    """Single measurement result used as an Operation parameter.

    A MeasuredParameter instance, given as a parameter to a
    :class:`~strawberryfields.ops.Operation` constructor, represents
    a dependence of the Operation on classical information obtained by
    measuring a subsystem of the register.

    Used for deferred measurements, i.e., using a measurement's value
    symbolically in defining a gate before the numeric value of that
    measurement is available.

    Former RegRefTransform (SF <= 0.11) functionality is provided by the sympy.Symbol base class.

    Args:
        regref (RegRef): register reference responsible for storing the measurement result
    """

    def __new__(cls, regref):
        # sympy.Basic.__new__ wants a name, other arguments must not end up in self._args
        return super().__new__(cls, "q" + str(regref.ind))

    def __init__(self, regref):
        if not regref.active:
            raise ValueError("Trying to use an inactive RegRef.")
        #: RegRef: the value of the parameter depends on this RegRef, and can only be evaluated after the corresponding subsystem has been measured
        self.regref = regref

    def _sympystr(self, printer):
        """Blackbird notation.

        The Sympy printing system uses this method instead of __str__.
        """
        return "q{}".format(self.regref.ind)

    def _eval_evalf(self, prec):
        """Returns the numeric result of the measurement if it is available.

        Returns:
            Any: measurement result

        Raises:
            ParameterError: iff the parameter has not been measured yet
        """
        res = self.regref.val
        if res is None:
            raise ParameterError(
                "{}: trying to use a nonexistent measurement result (e.g., before it has been measured).".format(
                    self
                )
            )
        return res


class FreeParameter(sympy.Symbol):
    """Named symbolic Operation parameter.

    Args:
        name (str): name of the free parameter
    """

    def __init__(self, name):
        #: str: name of the free parameter
        self.name = name
        #: Any: value of the parameter, None means unbound
        self.val = None
        #: Any: default value of the parameter, used if unbound
        self.default = None

    def _sympystr(self, printer):
        """Blackbird notation.

        The Sympy printing system uses this method instead of __str__.
        """
        return "{{{}}}".format(self.name)

    def _eval_evalf(self, prec):
        """Returns the value of the parameter if it has been bound, or the default value if not.

        Returns:
            Any: bound value, or the default value if not bound

        Raises:
            ParameterError: iff the parameter has not been bound, and has no default value
        """
        if self.val is None:
            if self.default is None:
                raise ParameterError("{}: unbound parameter with no default value.".format(self))
            return self.default
        return self.val



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *
from strawberryfields.utils import scale
from numpy import pi, sqrt
import numpy as np

# initialize engine and program objects
eng = sf.Engine(backend="gaussian")
gaussian_cloning = sf.Program(4)

with gaussian_cloning.context as q:
    # state to be cloned
    Coherent(0.7+1.2j) | q[0]

    # 50-50 beamsplitter
    BS = BSgate(pi/4, 0)

    # symmetric Gaussian cloning scheme
    BS | (q[0], q[1])
    BS | (q[1], q[2])
    MeasureX | q[1]
    MeasureP | q[2]
    Xgate(scale(q[1], sqrt(2))) | q[0]
    Zgate(scale(q[2], sqrt(2))) | q[0]

    # after the final beamsplitter, modes q[0] and q[3]
    # will contain identical approximate clones of the
    # initial state Coherent(0.1+0j)
    BS | (q[0], q[3])
    # end circuit

# run the engine
results = eng.run(gaussian_cloning, modes=[0, 3])

# return the cloning fidelity
fidelity = sqrt(results.state.fidelity_coherent([0.7+1.2j, 0.7+1.2j]))
# return the cloned displacement
alpha = results.state.displacement()

# run the engine over an ensemble
reps = 1000
f = np.empty([reps])
a = np.empty([reps], dtype=np.complex128)

for i in range(reps):
    eng.reset()
    results = eng.run(gaussian_cloning, modes=[0])
    f[i] = results.state.fidelity_coherent([0.7+1.2j])
    a[i] = results.state.displacement()

print("Fidelity of cloned state:", np.mean(f))
print("Mean displacement of cloned state:", np.mean(a))
print("Mean covariance matrix of cloned state:", np.cov([a.real, a.imag]))
